# coding=utf-8
"""
safety.py - Alerts about malicious URLs
Copyright © 2014, Elad Alfassa, <elad@fedoraproject.org>
Licensed under the Eiffel Forum License 2.

This plugin uses virustotal.com
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os.path
import re
import sys
import threading
import time

import requests

from sopel import formatting, plugin, tools
from sopel.config import types

try:
    # This is done separately from the below version if/else because JSONDecodeError
    # didn't appear until Python 3.5, but Sopel claims support for 3.3+
    # Redo this whole block of nonsense when dropping py2/old py3 support
    from json import JSONDecodeError as InvalidJSONResponse
except ImportError:
    InvalidJSONResponse = ValueError

if sys.version_info.major > 2:
    unicode = str
    from urllib.request import urlretrieve
    from urllib.parse import urlparse
else:
    from urllib import urlretrieve
    from urlparse import urlparse


LOGGER = logging.getLogger(__name__)
PLUGIN_OUTPUT_PREFIX = '[safety] '

vt_base_api_url = 'https://www.virustotal.com/vtapi/v2/url/'
malware_domains = set()
known_good = []
cache_limit = 512


class SafetySection(types.StaticSection):
    enabled_by_default = types.BooleanAttribute('enabled_by_default', default=True)
    """Whether to enable URL safety in all channels where it isn't explicitly disabled."""
    known_good = types.ListAttribute('known_good')
    """List of "known good" domains to ignore."""
    vt_api_key = types.ValidatedAttribute('vt_api_key')
    """Optional VirusTotal API key (improves malicious URL detection)."""


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | enabled\\_by\\_default | True | Enable URL safety in all channels where it isn't explicitly disabled. |
    | known\\_good | sopel.chat,dftba.net | List of "known good" domains to ignore. |
    | vt\\_api\\_key | 0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef | Optional VirusTotal API key to improve malicious URL detection |
    """
    config.define_section('safety', SafetySection)
    config.safety.configure_setting(
        'enabled_by_default',
        "Enable URL safety in channels that don't specifically disable it?",
    )
    config.safety.configure_setting(
        'known_good',
        'Enter any domains to whitelist',
    )
    config.safety.configure_setting(
        'vt_api_key',
        "Optionally, enter a VirusTotal API key to improve malicious URL "
        "protection.\nOtherwise, only the StevenBlack list will be used."
    )


def setup(bot):
    bot.config.define_section('safety', SafetySection)

    if 'safety_cache' not in bot.memory:
        bot.memory['safety_cache'] = tools.SopelMemory()
    if 'safety_cache_lock' not in bot.memory:
        bot.memory['safety_cache_lock'] = threading.Lock()
    for item in bot.config.safety.known_good:
        known_good.append(re.compile(item, re.I))

    old_file = os.path.join(bot.config.homedir, 'malwaredomains.txt')
    if os.path.exists(old_file) and os.path.isfile(old_file):
        LOGGER.info('Removing old malwaredomains file from %s', old_file)
        try:
            os.remove(old_file)
        except Exception as err:
            # for lack of a more specific error type...
            # Python on Windows throws an exception if the file is in use
            LOGGER.info('Could not delete %s: %s', old_file, str(err))

    loc = os.path.join(bot.config.homedir, 'unsafedomains.txt')
    if os.path.isfile(loc):
        if os.path.getmtime(loc) < time.time() - 24 * 60 * 60:
            # File exists but older than one day — update it
            _download_domain_list(loc)
    else:
        _download_domain_list(loc)
    with open(loc, 'r') as f:
        for line in f:
            if sys.version_info.major < 3:
                line = unicode(line, 'utf-8')
            else:
                line = unicode(line)

            clean_line = line.strip().lower()

            if not clean_line or clean_line[0] == '#':
                # blank line or comment
                continue

            parts = clean_line.split(' ', 1)
            try:
                domain = parts[1]
            except IndexError:
                # line does not contain a hosts entry; skip it
                continue

            if '.' in domain:
                # only publicly routable domains matter; skip loopback/link-local stuff
                malware_domains.add(domain)


def shutdown(bot):
    bot.memory.pop('safety_cache', None)
    bot.memory.pop('safety_cache_lock', None)


def _download_domain_list(path):
    url = 'https://raw.githubusercontent.com/StevenBlack/hosts/master/hosts'
    LOGGER.info('Downloading malicious domain list from %s', url)
    urlretrieve(url, path)


@plugin.rule(r'(?u).*(https?://\S+).*')
@plugin.priority('high')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def url_handler(bot, trigger):
    """Checks for malicious URLs"""
    check = True    # Enable URL checking
    strict = False  # Strict mode: kick on malicious URL
    positives = 0   # Number of engines saying it's malicious
    total = 0       # Number of total engines
    use_vt = True   # Use VirusTotal
    check = bot.config.safety.enabled_by_default
    if check is None:
        # If not set, assume default
        check = True
    # DB overrides config:
    setting = bot.db.get_channel_value(trigger.sender, 'safety')
    if setting is not None:
        if setting == 'off':
            return  # Not checking
        elif setting in ['on', 'strict', 'local', 'local strict']:
            check = True
        if setting == 'strict' or setting == 'local strict':
            strict = True
        if setting == 'local' or setting == 'local strict':
            use_vt = False

    if not check:
        return  # Not overridden by DB, configured default off

    try:
        netloc = urlparse(trigger.group(1)).netloc
    except ValueError:
        return  # Invalid IPv6 URL

    if any(regex.search(netloc) for regex in known_good):
        return  # Whitelisted

    apikey = bot.config.safety.vt_api_key
    try:
        if apikey is not None and use_vt:
            payload = {'resource': unicode(trigger),
                       'apikey': apikey,
                       'scan': '1'}

            if trigger not in bot.memory['safety_cache']:
                r = requests.post(vt_base_api_url + 'report', data=payload)
                r.raise_for_status()
                result = r.json()
                fetched = time.time()
                if all(k in result for k in ['positives', 'total']):
                    # cache result only if it contains a scan report
                    # TODO: handle checking back for results from queued scans
                    data = {'positives': result['positives'],
                            'total': result['total'],
                            'fetched': fetched}
                    bot.memory['safety_cache'][trigger] = data
                    if len(bot.memory['safety_cache']) >= (2 * cache_limit):
                        _clean_cache(bot)
            else:
                LOGGER.debug('using cache')
                result = bot.memory['safety_cache'][trigger]
            positives = result.get('positives', 0)
            total = result.get('total', 0)
    except requests.exceptions.RequestException:
        # Ignoring exceptions with VT so domain list will always work
        LOGGER.debug('[VirusTotal] Error obtaining response.', exc_info=True)
    except InvalidJSONResponse:
        # Ignoring exceptions with VT so domain list will always work
        LOGGER.debug('[VirusTotal] Malformed response (invalid JSON).', exc_info=True)

    if unicode(netloc).lower() in malware_domains:
        positives += 1
        total += 1

    if positives >= 1:
        # Possibly malicious URL detected!
        confidence = '{}%'.format(round((positives / total) * 100))
        msg = (
            'link posted by %s is possibly malicious '
            % formatting.bold(trigger.nick)
        )
        msg += '(confidence %s - %s/%s)' % (confidence, positives, total)
        warning = formatting.bold(formatting.color('WARNING:', 'red'))
        bot.say(warning + ' ' + msg)
        if strict:
            bot.kick(trigger.nick, trigger.sender, 'Posted a malicious link')


@plugin.command('safety')
@plugin.output_prefix(PLUGIN_OUTPUT_PREFIX)
def toggle_safety(bot, trigger):
    """Set safety setting for channel"""
    if not trigger.admin and bot.channels[trigger.sender].privileges[trigger.nick] < plugin.OP:
        bot.reply('Only channel operators can change safety settings')
        return
    allowed_states = ['strict', 'on', 'off', 'local', 'local strict']
    if not trigger.group(2) or trigger.group(2).lower() not in allowed_states:
        options = ' / '.join(allowed_states)
        bot.reply('Available options: %s' % options)
        return

    channel = trigger.sender.lower()
    bot.db.set_channel_value(channel, 'safety', trigger.group(2).lower())
    bot.say('Safety is now set to "%s" on this channel' % trigger.group(2))


# Clean the cache every day
# Code above also calls this if there are too many cache entries
@plugin.interval(24 * 60 * 60)
def _clean_cache(bot):
    """Cleans up old entries in URL safety cache."""
    if bot.memory['safety_cache_lock'].acquire(False):
        LOGGER.info('Starting safety cache cleanup...')
        try:
            # clean up by age first
            cutoff = time.time() - (7 * 24 * 60 * 60)  # 7 days ago
            old_keys = []
            for key, data in tools.iteritems(bot.memory['safety_cache']):
                if data['fetched'] <= cutoff:
                    old_keys.append(key)
            for key in old_keys:
                bot.memory['safety_cache'].pop(key, None)

            # clean up more values if the cache is still too big
            overage = len(bot.memory['safety_cache']) - cache_limit
            if overage > 0:
                extra_keys = sorted(
                    (data.fetched, key)
                    for (key, data)
                    in bot.memory['safety_cache'].items())[:overage]
                for (_, key) in extra_keys:
                    bot.memory['safety_cache'].pop(key, None)
        finally:
            # No matter what errors happen (or not), release the lock
            bot.memory['safety_cache_lock'].release()

        LOGGER.info('Safety cache cleanup finished.')
    else:
        LOGGER.info(
            'Skipping safety cache cleanup: Cache is locked, '
            'cleanup already running.')



# coding=utf-8
"""Sopel's plugin rules management.

.. versionadded:: 7.1

.. important::

    This is all fresh and new. Its usage and documentation is for Sopel core
    development and advanced developers. It is subject to rapid changes
    between versions without much (or any) warning.

    Do **not** build your plugin based on what is here, you do **not** need to.

"""
# Copyright 2020, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals


import datetime
import functools
import inspect
import itertools
import logging
import re
import threading


from sopel import tools
from sopel.config.core_section import (
    COMMAND_DEFAULT_HELP_PREFIX, COMMAND_DEFAULT_PREFIX, URL_DEFAULT_SCHEMES)


try:
    from urllib.parse import urlparse
except ImportError:
    # TODO: remove when dropping Python 2.7
    from urlparse import urlparse

try:
    from inspect import getfullargspec as inspect_getargspec
except ImportError:
    # TODO: remove when dropping Python 2.7
    from inspect import getargspec as inspect_getargspec


__all__ = [
    'Manager',
    'Rule',
    'FindRule',
    'SearchRule',
    'Command',
    'NickCommand',
    'ActionCommand',
    'URLCallback',
]


LOGGER = logging.getLogger(__name__)

IGNORE_RATE_LIMIT = 1  # equivalent to sopel.plugin.NOLIMIT
"""Return value used to indicate that rate-limiting should be ignored."""
PRIORITY_HIGH = 'high'
"""Highest rule priority."""
PRIORITY_MEDIUM = 'medium'
"""Medium rule priority."""
PRIORITY_LOW = 'low'
"""Lowest rule priority."""
PRIORITY_SCALES = {
    PRIORITY_HIGH: 0,
    PRIORITY_MEDIUM: 100,
    PRIORITY_LOW: 1000,
}
"""Mapping of priority label to priority scale."""

# Can be implementation-dependent
_regex_type = type(re.compile(''))


def _clean_rules(rules, nick, aliases):
    for pattern in rules:
        if isinstance(pattern, _regex_type):
            # already a compiled regex
            yield pattern
        else:
            yield _compile_pattern(pattern, nick, aliases)


def _compile_pattern(pattern, nick, aliases=None):
    if aliases:
        nicks = list(aliases)  # alias_nicks.copy() doesn't work in py2
        nicks.append(nick)
        nicks = map(re.escape, nicks)
        nick = '(?:%s)' % '|'.join(nicks)
    else:
        nick = re.escape(nick)

    pattern = pattern.replace('$nickname', nick)
    pattern = pattern.replace('$nick ', r'{}[,:]\s*'.format(nick))  # @rule('$nick hi')
    pattern = pattern.replace('$nick', r'{}[,:]\s+'.format(nick))  # @rule('$nickhi')
    flags = re.IGNORECASE
    if '\n' in pattern:
        # See https://docs.python.org/3/library/re.html#re.VERBOSE
        flags |= re.VERBOSE
    return re.compile(pattern, flags)


def _has_labeled_rule(registry, label, plugin=None):
    rules = (
        itertools.chain(*registry.values())
        if plugin is None
        else registry.get(plugin, [])
    )
    return any(rule.get_rule_label() == label for rule in rules)


def _has_named_rule(registry, name, follow_alias=False, plugin=None):
    rules = registry.values() if plugin is None else [registry.get(plugin, {})]

    has_name = any(
        (name in plugin_rules)
        for plugin_rules in rules
    )
    aliases = (
        rule.has_alias(name)
        for plugin_rules in rules
        for rule in plugin_rules.values()
    )

    return has_name or (follow_alias and any(aliases))


def _clean_callable_examples(examples):
    valid_keys = [
        # message
        'example',
        'result',
        # flags
        'is_private_message',
        'is_help',
        'is_pattern',
        'is_admin',
        'is_owner',
    ]

    return tuple(
        dict(
            (key, value)
            for key, value in example.items()
            if key in valid_keys
        )
        for example in examples
    )


class Manager(object):
    """Manager of plugin rules.

    This manager stores plugin rules and can then provide the matching rules
    for a given trigger.

    To register a rule:

    * :meth:`register` for generic rules
    * :meth:`register_command` for named rules with a prefix
    * :meth:`register_nick_command` for named rules based on nick calling
    * :meth:`register_action_command` for named rules based on ``ACTION``
    * :meth:`register_url_callback` for URL callback rules

    Then to match the rules against a ``trigger``, see the
    :meth:`get_triggered_rules`, which returns a list of ``(rule, match)``,
    sorted by priorities (high first, medium second, and low last).
    """
    def __init__(self):
        self._rules = tools.SopelMemoryWithDefault(list)
        self._commands = tools.SopelMemoryWithDefault(dict)
        self._nick_commands = tools.SopelMemoryWithDefault(dict)
        self._action_commands = tools.SopelMemoryWithDefault(dict)
        self._url_callbacks = tools.SopelMemoryWithDefault(list)
        self._register_lock = threading.Lock()

    def unregister_plugin(self, plugin_name):
        """Unregister all the rules from a plugin.

        :param str plugin_name: the name of the plugin to remove
        :return: the number of rules unregistered for this plugin
        :rtype: int

        All rules, commands, nick commands, and action commands of that plugin
        will be removed from the manager.
        """
        registries = [
            self._rules,
            self._commands,
            self._nick_commands,
            self._action_commands,
            self._url_callbacks,
        ]

        unregistered_rules = 0
        with self._register_lock:
            for registry in registries:
                rules_count = len(registry[plugin_name])
                del registry[plugin_name]
                unregistered_rules = unregistered_rules + rules_count

        LOGGER.debug(
            '[%s] Successfully unregistered %d rules',
            plugin_name,
            unregistered_rules)

        return unregistered_rules

    def register(self, rule):
        """Register a plugin rule.

        :param rule: the rule to register
        :type rule: :class:`Rule`
        """
        with self._register_lock:
            self._rules[rule.get_plugin_name()].append(rule)
        LOGGER.debug('Rule registered: %s', str(rule))

    def register_command(self, command):
        """Register a plugin command.

        :param command: the command to register
        :type command: :class:`Command`
        """
        with self._register_lock:
            plugin = command.get_plugin_name()
            self._commands[plugin][command.name] = command
        LOGGER.debug('Command registered: %s', str(command))

    def register_nick_command(self, command):
        """Register a plugin nick command.

        :param command: the nick command to register
        :type command: :class:`NickCommand`
        """
        with self._register_lock:
            plugin = command.get_plugin_name()
            self._nick_commands[plugin][command.name] = command
        LOGGER.debug('Nick Command registered: %s', str(command))

    def register_action_command(self, command):
        """Register a plugin action command.

        :param command: the action command to register
        :type command: :class:`ActionCommand`
        """
        with self._register_lock:
            plugin = command.get_plugin_name()
            self._action_commands[plugin][command.name] = command
        LOGGER.debug('Action Command registered: %s', str(command))

    def register_url_callback(self, url_callback):
        """Register a plugin URL callback.

        :param url_callback: the URL callback to register
        :type url_callback: :class:`URLCallback`
        """
        with self._register_lock:
            plugin = url_callback.get_plugin_name()
            self._url_callbacks[plugin].append(url_callback)
        LOGGER.debug('URL callback registered: %s', str(url_callback))

    def has_rule(self, label, plugin=None):
        """Tell if the manager knows a rule with this ``label``.

        :param str label: the label of the rule to look for
        :param str plugin: optional filter on the plugin name
        :return: ``True`` if the rule exists, ``False`` otherwise
        :rtype: bool

        The optional parameter ``plugin`` can be provided to limit the rules
        to only those from that plugin.
        """
        return _has_labeled_rule(self._rules, label, plugin)

    def has_command(self, name, follow_alias=True, plugin=None):
        """Tell if the manager knows a command with this ``name``.

        :param str label: the label of the rule to look for
        :param bool follow_alias: optional flag to include aliases
        :param str plugin: optional filter on the plugin name
        :return: ``True`` if the command exists, ``False`` otherwise
        :rtype: bool

        By default, this method follows aliases to search commands. If the
        optional parameter ``follow_alias`` is ``False``, then it won't find
        commands by their aliases::

            >>> command = Command('hi', prefix='"', aliases=['hey'])
            >>> manager.register_command(command)
            >>> manager.has_command('hi')
            True
            >>> manager.has_command('hey')
            True
            >>> manager.has_command('hey', follow_alias=False)
            False

        The optional parameter ``plugin`` can be provided to limit the commands
        to the ones of said plugin.
        """
        return _has_named_rule(self._commands, name, follow_alias, plugin)

    def has_nick_command(self, name, follow_alias=True, plugin=None):
        """Tell if the manager knows a nick command with this ``name``.

        :param str label: the label of the rule to look for
        :param bool follow_alias: optional flag to include aliases
        :param str plugin: optional filter on the plugin name
        :return: ``True`` if the command exists, ``False`` otherwise
        :rtype: bool

        This method works like :meth:`has_command`, but with nick commands.
        """
        return _has_named_rule(self._nick_commands, name, follow_alias, plugin)

    def has_action_command(self, name, follow_alias=True, plugin=None):
        """Tell if the manager knows an action command with this ``name``.

        :param str label: the label of the rule to look for
        :param bool follow_alias: optional flag to include aliases
        :param str plugin: optional filter on the plugin name
        :return: ``True`` if the command exists, ``False`` otherwise
        :rtype: bool

        This method works like :meth:`has_command`, but with action commands.
        """
        return _has_named_rule(
            self._action_commands, name, follow_alias, plugin)

    def has_url_callback(self, label, plugin=None):
        """Tell if the manager knows a URL callback with this ``label``.

        :param str label: the label of the URL callback to look for
        :param str plugin: optional filter on the plugin name
        :return: ``True`` if the URL callback exists, ``False`` otherwise
        :rtype: bool

        The optional parameter ``plugin`` can be provided to limit the URL
        callbacks to only those from that plugin.
        """
        return _has_labeled_rule(self._url_callbacks, label, plugin)

    def get_all_commands(self):
        """Retrieve all the registered commands, by plugin.

        :return: a list of 2-value tuples as ``(key, value)``, where each key
                 is a plugin name, and the value is a ``dict`` of its
                 :term:`commands <Command>`
        """
        # expose a copy of the registered commands
        return self._commands.items()

    def get_all_nick_commands(self):
        """Retrieve all the registered nick commands, by plugin.

        :return: a list of 2-value tuples as ``(key, value)``, where each key
                 is a plugin name, and the value is a ``dict`` of its
                 :term:`nick commands <Nick command>`
        """
        # expose a copy of the registered commands
        return self._nick_commands.items()

    def get_all_action_commands(self):
        """Retrieve all the registered action commands, by plugin.

        :return: a list of 2-value tuples as ``(key, value)``, where each key
                 is a plugin name, and the value is a ``dict`` of its
                 :term:`action commands <Action command>`
        """
        # expose a copy of the registered action commands
        return self._action_commands.items()

    def get_all_generic_rules(self):
        """Retrieve all the registered generic rules, by plugin.

        :return: a list of 2-value tuples as ``(key, value)``, where each key
                 is a plugin name, and the value is a ``list`` of its
                 :term:`generic rules <Generic rule>`
        """
        # expose a copy of the registered generic rules
        return self._rules.items()

    def get_all_url_callbacks(self):
        """Retrieve all the registered URL callbacks, by plugin.

        :return: a list of 2-value tuples as ``(key, value)``, where each key
                 is a plugin name, and the value is a ``list`` of its
                 :term:`URL callbacks <URL callback>`
        """
        # expose a copy of the registered generic rules
        return self._url_callbacks.items()

    def get_triggered_rules(self, bot, pretrigger):
        """Get triggered rules with their match objects, sorted by priorities.

        :param bot: Sopel instance
        :type bot: :class:`sopel.bot.Sopel`
        :param pretrigger: IRC line
        :type pretrigger: :class:`sopel.trigger.PreTrigger`
        :return: a tuple of ``(rule, match)``, sorted by priorities
        :rtype: tuple
        """
        generic_rules = self._rules.values()
        command_rules = (
            rules_dict.values()
            for rules_dict in self._commands.values())
        nick_rules = (
            rules_dict.values()
            for rules_dict in self._nick_commands.values())
        action_rules = (
            rules_dict.values()
            for rules_dict in self._action_commands.values())
        url_callback_rules = self._url_callbacks.values()

        rules = itertools.chain(
            itertools.chain(*generic_rules),
            itertools.chain(*command_rules),
            itertools.chain(*nick_rules),
            itertools.chain(*action_rules),
            itertools.chain(*url_callback_rules),
        )
        matches = (
            (rule, match)
            for rule in rules
            for match in rule.match(bot, pretrigger)
        )
        # Returning a tuple instead of a sorted object ensures that:
        #   1. it's not a lazy object
        #   2. it's an immutable iterable
        # We can't accept lazy evaluation or yield results; it has to be a
        # static list of (rule/match), otherwise Python will raise an error
        # if any rule execution tries to alter the list of registered rules.
        # Making it immutable is the cherry on top.
        return tuple(sorted(matches, key=lambda x: x[0].priority_scale))

    def check_url_callback(self, bot, url):
        """Tell if the ``url`` matches any of the registered URL callbacks.

        :param bot: Sopel instance
        :type bot: :class:`sopel.bot.Sopel`
        :param str url: URL to check
        :return: ``True`` when ``url`` matches any URL callbacks,
                 ``False`` otherwise
        :rtype: bool
        """
        return any(
            any(rule.parse(url))
            for plugin_rules in self._url_callbacks.values()
            for rule in plugin_rules
        )


class AbstractRule(object):
    """Abstract definition of a plugin's rule.

    Any rule class must be an implementation of this abstract class, as it
    defines the Rule interface:

    * plugin name
    * priority
    * label
    * doc, usages, and tests
    * output prefix
    * matching patterns, events, and intents
    * allow echo-message
    * threaded execution or not
    * rate limiting feature
    * text parsing
    * and finally, trigger execution (i.e. actually doing something)

    """
    @classmethod
    def from_callable(cls, settings, handler):
        """Instantiate a rule object from ``settings`` and ``handler``.

        :param settings: Sopel's settings
        :type settings: :class:`sopel.config.Config`
        :param callable handler: a function-based rule handler
        :return: an instance of this class created from the ``handler``
        :rtype: :class:`AbstractRule`

        Sopel's function-based rule handlers are simple callables, decorated
        with :mod:`sopel.plugin`'s decorators to add attributes, such as rate
        limit, threaded execution, output prefix, priority, and so on. In order
        to load these functions as rule objects, this class method can be used;
        it takes the bot's ``settings`` and a cleaned ``handler``.

        A "cleaned handler" is a function, decorated appropriately, and passed
        through the filter of the
        :func:`loader's clean<sopel.loader.clean_callable>` function.
        """
        raise NotImplementedError

    @property
    def priority_scale(self):
        """Rule's priority on a numeric scale.

        This attribute can be used to sort rules between each other, the
        highest priority rules coming first. The default priority for a rule
        is "medium".
        """
        priority_key = self.get_priority()

        return (
            PRIORITY_SCALES.get(priority_key) or
            PRIORITY_SCALES[PRIORITY_MEDIUM]
        )

    def get_plugin_name(self):
        """Get the rule's plugin name.

        :rtype: str

        The rule's plugin name will be used in various places to select,
        register, unregister, and manipulate the rule based on its plugin,
        which is referenced by its name.
        """
        raise NotImplementedError

    def get_rule_label(self):
        """Get the rule's label.

        :rtype: str

        A rule can have a label, which can identify the rule by string, the
        same way a plugin can be identified by its name. This label can be used
        to select, register, unregister, and manipulate the rule based on its
        own label. Note that the label has no effect on the rule's execution.
        """
        raise NotImplementedError

    def get_usages(self):
        """Get the rule's usage examples.

        :rtype: tuple

        A rule can have usage examples, i.e. a list of examples showing how
        the rule can be used, or in what context it can be triggered.
        """
        raise NotImplementedError

    def get_test_parameters(self):
        """Get parameters for automated tests.

        :rtype: tuple

        A rule can have automated tests attached to it, and this method must
        return the test parameters:

        * the expected IRC line
        * the expected line of results, as said by the bot
        * if the user should be an admin or not
        * if the results should be used as regex pattern

        .. seealso::

            :meth:`sopel.plugin.example` for more about test parameters.

        """
        raise NotImplementedError

    def get_doc(self):
        """Get the rule's documentation.

        :rtype: str

        A rule's documentation is a short text that can be displayed to a user
        on IRC upon asking for help about this rule. The equivalent of Python
        docstrings, but for IRC rules.
        """
        raise NotImplementedError

    def get_priority(self):
        """Get the rule's priority.

        :rtype: str

        A rule can have a priority, based on the three pre-defined priorities
        used by Sopel: ``PRIORITY_HIGH``, ``PRIORITY_MEDIUM``, and
        ``PRIORITY_LOW``.

        .. seealso::

            The :attr:`AbstractRule.priority_scale` property uses this method
            to look up the numeric priority value, which is used to sort rules
            by priority.

        """
        raise NotImplementedError

    def get_output_prefix(self):
        """Get the rule's output prefix.

        :rtype: str

        .. seealso::

            See the :class:`sopel.bot.SopelWrapper` class for more information
            on how the output prefix can be used.
        """
        raise NotImplementedError

    def match(self, bot, pretrigger):
        """Match a pretrigger according to the rule.

        :param bot: Sopel instance
        :type bot: :class:`sopel.bot.Sopel`
        :param pretrigger: line to match
        :type pretrigger: :class:`sopel.trigger.PreTrigger`

        This method must return a list of `match objects`__.

        .. __: https://docs.python.org/3.6/library/re.html#match-objects
        """
        raise NotImplementedError

    def match_event(self, event):
        """Tell if the rule matches this ``event``.

        :param str event: potential matching event
        :return: ``True`` when ``event`` matches the rule, ``False`` otherwise
        :rtype: bool
        """
        raise NotImplementedError

    def match_intent(self, intent):
        """Tell if the rule matches this ``intent``.

        :param str intent: potential matching intent
        :return: ``True`` when ``intent`` matches the rule, ``False`` otherwise
        :rtype: bool
        """
        raise NotImplementedError

    def allow_echo(self):
        """Tell if the rule should match echo messages.

        :return: ``True`` when the rule allows echo messages,
                 ``False`` otherwise
        :rtype: bool
        """
        raise NotImplementedError

    def is_threaded(self):
        """Tell if the rule's execution should be in a thread.

        :return: ``True`` if the execution should be in a thread,
                 ``False`` otherwise
        :rtype: bool
        """
        raise NotImplementedError

    def is_unblockable(self):
        """Tell if the rule is unblockable.

        :return: ``True`` when the rule is unblockable, ``False`` otherwise
        :rtype: bool
        """
        raise NotImplementedError

    def is_rate_limited(self, nick):
        """Tell when the rule reached the ``nick``'s rate limit.

        :return: ``True`` when the rule reached the limit, ``False`` otherwise
        :rtype: bool
        """
        raise NotImplementedError

    def is_channel_rate_limited(self, channel):
        """Tell when the rule reached the ``channel``'s rate limit.

        :return: ``True`` when the rule reached the limit, ``False`` otherwise
        :rtype: bool
        """
        raise NotImplementedError

    def is_global_rate_limited(self):
        """Tell when the rule reached the server's rate limit.

        :return: ``True`` when the rule reached the limit, ``False`` otherwise
        :rtype: bool
        """
        raise NotImplementedError

    def parse(self, text):
        """Parse ``text`` and yield matches.

        :param str text: text to parse by the rule
        :return: yield a list of match object
        :rtype: generator of `re.match`__

        .. __: https://docs.python.org/3.6/library/re.html#match-objects
        """
        raise NotImplementedError

    def execute(self, bot, trigger):
        """Execute the triggered rule.

        :param bot: Sopel wrapper
        :type bot: :class:`sopel.bot.SopelWrapper`
        :param trigger: IRC line
        :type trigger: :class:`sopel.trigger.Trigger`

        This is the method called by the bot when a rule matches a ``trigger``.
        """
        raise NotImplementedError


class Rule(AbstractRule):
    """Generic rule definition.

    A generic rule (or simply "a rule") uses regular expressions to match
    at most once per IRC line per regular expression, i.e. you can trigger
    between 0 and the number of regex the rule has per IRC line.

    Here is an example with a rule with the pattern ``r'hello (\\w+)'``:

    .. code-block:: irc

        <user> hello here
        <Bot> You triggered a rule, saying hello to "here"
        <user> hello sopelunkers
        <Bot> You triggered a rule, saying hello to "sopelunkers"

    Generic rules are not triggered by any specific name, unlike commands which
    have names and aliases.
    """

    REGEX_ATTRIBUTE = 'rule'
    LAZY_ATTRIBUTE = 'rule_lazy_loaders'

    @classmethod
    def kwargs_from_callable(cls, handler):
        """Generate the keyword arguments to create a new instance.

        :param callable handler: callable used to generate keyword arguments
        :return: a map of keyword arguments
        :rtype: dict

        This classmethod takes the ``handler``'s attributes to generate a map
        of keyword arguments for the class. This can be used by the
        :meth:`from_callable` classmethod to instantiate a new rule object.

        The expected attributes are the ones set by decorators from the
        :mod:`sopel.plugin` module.
        """
        # manage examples:
        # - usages are for documentation only
        # - tests are examples that can be run and tested
        examples = _clean_callable_examples(
            getattr(handler, 'example', None) or tuple())

        usages = tuple(
            example
            for example in examples
            if example.get('is_help')
        ) or examples and (examples[0],)
        tests = tuple(example for example in examples if example.get('result'))

        return {
            'plugin': getattr(handler, 'plugin_name', None),
            'label': getattr(handler, 'rule_label', None),
            'priority': getattr(handler, 'priority', PRIORITY_MEDIUM),
            'events': getattr(handler, 'event', []),
            'intents': getattr(handler, 'intents', []),
            'allow_echo': getattr(handler, 'echo', False),
            'threaded': getattr(handler, 'thread', True),
            'output_prefix': getattr(handler, 'output_prefix', ''),
            'unblockable': getattr(handler, 'unblockable', False),
            'rate_limit': getattr(handler, 'rate', 0),
            'channel_rate_limit': getattr(handler, 'channel_rate', 0),
            'global_rate_limit': getattr(handler, 'global_rate', 0),
            'usages': usages or tuple(),
            'tests': tests,
            'doc': inspect.getdoc(handler),
        }

    @classmethod
    def regex_from_callable(cls, settings, handler):
        regexes = getattr(handler, cls.REGEX_ATTRIBUTE, []) or []
        if not regexes:
            raise RuntimeError(
                'Invalid rule handler: %s' % handler)

        return tuple(_clean_rules(
            regexes,
            settings.core.nick,
            settings.core.alias_nicks,
        ))

    @classmethod
    def regex_from_callable_lazy(cls, settings, handler):
        lazy_loaders = getattr(handler, cls.LAZY_ATTRIBUTE, [])
        if not lazy_loaders:
            raise RuntimeError(
                'Invalid lazy rule: %s' % handler)

        loader = tools.chain_loaders(*lazy_loaders)
        regexes = loader(settings)

        if not regexes:
            raise RuntimeError(
                'Invalid lazy loader: %s' % handler)

        return regexes

    @classmethod
    def from_callable(cls, settings, handler):
        regexes = cls.regex_from_callable(settings, handler)
        kwargs = cls.kwargs_from_callable(handler)
        kwargs['handler'] = handler

        return cls(regexes, **kwargs)

    @classmethod
    def from_callable_lazy(cls, settings, handler):
        """Instantiate a rule object from a handler with lazy-loaded regexes.

        :param settings: Sopel's settings
        :type settings: :class:`sopel.config.Config`
        :param callable handler: a function-based rule handler with a
                                 lazy-loader for the regexes
        :return: an instance of this class created from the ``handler``
        :rtype: :class:`AbstractRule`

        Similar to the :meth:`from_callable` classmethod, it requires a rule
        handler decorated with :mod:`sopel.plugin`'s decorators.

        Unlike the :meth:`from_callable` classmethod, the regexes are not
        already attached to the handler: its loader functions will be used to
        get the rule's regexes. See the :func:`sopel.plugin.rule_lazy`
        decorator for more information about the handler and the loaders'
        signatures.

        .. seealso::

            The handler can have more than one loader attached. In that case,
            these loaders are chained with :func:`sopel.tools.chain_loaders`.

        """
        regexes = cls.regex_from_callable_lazy(settings, handler)
        kwargs = cls.kwargs_from_callable(handler)
        kwargs['handler'] = handler

        return cls(regexes, **kwargs)

    def __init__(self,
                 regexes,
                 plugin=None,
                 label=None,
                 priority=PRIORITY_MEDIUM,
                 handler=None,
                 events=None,
                 intents=None,
                 allow_echo=False,
                 threaded=True,
                 output_prefix=None,
                 unblockable=False,
                 rate_limit=0,
                 channel_rate_limit=0,
                 global_rate_limit=0,
                 usages=None,
                 tests=None,
                 doc=None):
        # core
        self._regexes = regexes
        self._plugin_name = plugin
        self._label = label
        self._priority = priority or PRIORITY_MEDIUM
        self._handler = handler

        # filters
        self._events = events or ['PRIVMSG']
        self._intents = intents or []
        self._allow_echo = bool(allow_echo)

        # execution
        self._threaded = bool(threaded)
        self._output_prefix = output_prefix or ''

        # rate limiting
        self._unblockable = bool(unblockable)
        self._rate_limit = rate_limit
        self._channel_rate_limit = channel_rate_limit
        self._global_rate_limit = global_rate_limit

        # metrics
        self._metrics_nick = {}
        self._metrics_sender = {}
        self._metrics_global = None

        # docs & tests
        self._usages = usages or tuple()
        self._tests = tests or tuple()
        self._doc = doc or ''

    def __str__(self):
        try:
            label = self.get_rule_label()
        except RuntimeError:
            label = '(generic)'

        plugin = self.get_plugin_name() or '(no-plugin)'

        return '<%s %s.%s (%d)>' % (
            self.__class__.__name__, plugin, label, len(self._regexes))

    def get_plugin_name(self):
        return self._plugin_name

    def get_rule_label(self):
        """Get the rule's label.

        :rtype: str
        :raise RuntimeError: when the label is undefined

        Return its label if it has one, or the value of its ``handler``'s
        ``__name__``, if it has a handler. If both methods fail, a
        :exc:`RuntimeError` is raised because the rule has an undefined label.
        """
        if self._label:
            return self._label

        if self._handler and self._handler.__name__:
            return self._handler.__name__

        raise RuntimeError('Undefined rule label')

    def get_usages(self):
        return tuple(
            {
                'text': usage['example'],
                'result': usage.get('result', None),
                'is_pattern': bool(usage.get('is_pattern')),
                'is_admin': bool(usage.get('is_admin')),
                'is_owner': bool(usage.get('is_owner')),
                'is_private_message': bool(usage.get('is_private_message')),
            }
            for usage in self._usages
            if usage.get('example')
        )

    def get_test_parameters(self):
        return self._tests or tuple()

    def get_doc(self):
        return self._doc

    def get_priority(self):
        return self._priority

    def get_output_prefix(self):
        return self._output_prefix

    def match(self, bot, pretrigger):
        args = pretrigger.args
        text = args[-1] if args else ''

        if not self.match_preconditions(bot, pretrigger):
            return []

        # parse text
        return self.parse(text)

    def match_preconditions(self, bot, pretrigger):
        event = pretrigger.event
        intent = pretrigger.tags.get('intent')
        nick = pretrigger.nick
        is_echo_message = (
            nick.lower() == bot.nick.lower() and
            event in ["PRIVMSG", "NOTICE"]
        )

        return (
            self.match_event(event) and
            self.match_intent(intent) and
            (not is_echo_message or self.allow_echo())
        )

    def parse(self, text):
        for regex in self._regexes:
            result = regex.match(text)
            if result:
                yield result

    def match_event(self, event):
        return bool(event and event in self._events)

    def match_intent(self, intent):
        if not self._intents:
            return True

        return bool(intent and any(
            regex.match(intent)
            for regex in self._intents
        ))

    def allow_echo(self):
        return self._allow_echo

    def is_threaded(self):
        return self._threaded

    def is_unblockable(self):
        return self._unblockable

    def is_rate_limited(self, nick):
        metrics = self._metrics_nick.get(nick)
        if metrics is None:
            return False
        last_usage_at, exit_code = metrics

        if exit_code == IGNORE_RATE_LIMIT:
            return False

        now = datetime.datetime.utcnow()
        rate_limit = datetime.timedelta(seconds=self._rate_limit)
        return (now - last_usage_at) <= rate_limit

    def is_channel_rate_limited(self, channel):
        metrics = self._metrics_sender.get(channel)
        if metrics is None:
            return False
        last_usage_at, exit_code = metrics

        if exit_code == IGNORE_RATE_LIMIT:
            return False

        now = datetime.datetime.utcnow()
        rate_limit = datetime.timedelta(seconds=self._channel_rate_limit)
        return (now - last_usage_at) <= rate_limit

    def is_global_rate_limited(self):
        metrics = self._metrics_global
        if metrics is None:
            return False
        last_usage_at, exit_code = metrics

        if exit_code == IGNORE_RATE_LIMIT:
            return False

        now = datetime.datetime.utcnow()
        rate_limit = datetime.timedelta(seconds=self._global_rate_limit)
        return (now - last_usage_at) <= rate_limit

    def execute(self, bot, trigger):
        if not self._handler:
            raise RuntimeError('Improperly configured rule: no handler')

        # execute the handler
        exit_code = self._handler(bot, trigger)

        # register metrics
        now = datetime.datetime.utcnow()
        self._metrics_nick[trigger.nick] = (now, exit_code)
        self._metrics_sender[trigger.sender] = (now, exit_code)
        self._metrics_global = (now, exit_code)

        # return exit code
        return exit_code


class NamedRuleMixin(object):
    """Mixin for named rules.

    A named rule is invoked by using a specific word, and is usually known
    as a "command". For example, the command "hello" is triggered by using
    the word "hello" with some sort of prefix or context.

    A named rule can be invoked by using one of its aliases, also.
    """
    @property
    def name(self):
        return self._name

    @property
    def aliases(self):
        return self._aliases

    def get_rule_label(self):
        """Get the rule's label.

        :rtype: str

        A named rule's label is its name.
        """
        return self._name.replace(' ', '-')

    def has_alias(self, name):
        """Tell when ``name`` is one of the rule's aliases.

        :param str name: potential alias name
        :return: ``True`` when ``name`` is an alias, ``False`` otherwise
        :rtype: bool
        """
        return name in self._aliases

    def escape_name(self, name):
        """Escape the provided name if needed.

        .. note::

            Until now, Sopel has allowed command name to be regex pattern.
            It was mentioned in the documentation without much details, and
            there were no tests for it.

            In order to ensure backward compatibility with previous versions of
            Sopel, we make sure to escape command name only when it's needed.

            **It is not recommended to use a regex pattern for your command
            name. This feature will be removed in Sopel 8.0.**

        """
        if set('.^$*+?{}[]\\|()') & set(name):
            # the name contains a regex pattern special character
            # we assume the user knows what they are doing
            try:
                # make sure it compiles properly
                # (nobody knows what they are doing)
                re.compile(name)
            except re.error as error:
                original_name = name
                name = re.escape(name)
                LOGGER.warning(
                    'Command name "%s" is an invalid regular expression '
                    'and will be replaced by "%s": %s',
                    original_name, name, error)
        else:
            name = re.escape(name)

        return name


class Command(NamedRuleMixin, Rule):
    """Command rule definition.

    A command rule (or simply "a command") is a named rule, i.e. it has a known
    name and must be invoked using that name (or one of its aliases, if any).
    Apart from that, it behaves exactly like a :class:`generic rule <Rule>`.

    Here is an example with the ``dummy`` command:

    .. code-block:: irc

        <user> .dummy
        <Bot> You just invoked the command 'dummy'
        <user> .dummy-alias
        <Bot> You just invoked the command 'dummy' (as 'dummy-alias')

    """
    # This regexp matches equivalently and produces the same
    # groups 1 and 2 as the old regexp: r'^%s(%s)(?: +(.*))?$'
    # The only differences should be handling all whitespace
    # like spaces and the addition of groups 3-6.
    PATTERN_TEMPLATE = r"""
        (?:{prefix})({command}) # Command as group 1.
        (?:\s+              # Whitespace to end command.
        (                   # Rest of the line as group 2.
        (?:(\S+))?          # Parameters 1-4 as groups 3-6.
        (?:\s+(\S+))?
        (?:\s+(\S+))?
        (?:\s+(\S+))?
        .*                  # Accept anything after the parameters.
                            # Leave it up to the plugin to parse the line.
        ))?                 # Group 2 must be None, if there are no parameters.
        $                   # EoL, so there are no partial matches.
    """

    @classmethod
    def from_callable(cls, settings, handler):
        prefix = settings.core.prefix
        help_prefix = settings.core.help_prefix
        commands = getattr(handler, 'commands', [])
        if not commands:
            raise RuntimeError('Invalid command callable: %s' % handler)

        name = commands[0]
        aliases = commands[1:]
        kwargs = cls.kwargs_from_callable(handler)
        kwargs.update({
            'name': name,
            'prefix': prefix,
            'help_prefix': help_prefix,
            'aliases': aliases,
            'handler': handler,
        })

        return cls(**kwargs)

    def __init__(self,
                 name,
                 prefix=COMMAND_DEFAULT_PREFIX,
                 help_prefix=COMMAND_DEFAULT_HELP_PREFIX,
                 aliases=None,
                 **kwargs):
        super(Command, self).__init__([], **kwargs)
        self._name = name
        self._prefix = prefix
        self._help_prefix = help_prefix
        self._aliases = tuple(aliases) if aliases is not None else tuple()
        self._regexes = (self.get_rule_regex(),)

    def __str__(self):
        label = self.get_rule_label()
        plugin = self.get_plugin_name() or '(no-plugin)'
        aliases = '|'.join(self._aliases)

        return '<Command %s.%s [%s]>' % (plugin, label, aliases)

    def get_usages(self):
        usages = []

        for usage in self._usages:
            text = usage.get('example')
            if not text:
                continue

            if text[0] != self._help_prefix:
                text = text.replace(
                    COMMAND_DEFAULT_HELP_PREFIX, self._help_prefix, 1)

            new_usage = {
                'text': text,
                'result': usage.get('result', None),
                'is_pattern': bool(usage.get('is_pattern')),
                'is_admin': bool(usage.get('is_admin')),
                'is_owner': bool(usage.get('is_owner')),
                'is_private_message': bool(usage.get('is_private_message')),
            }
            usages.append(new_usage)

        return tuple(usages)

    def get_rule_regex(self):
        """Make the rule regex for this command.

        :return: a compiled regex for this command and its aliases

        The command regex factors in:

        * the prefix regular expression,
        * the rule's name (escaped for regex if needed),
        * all of its aliases (escaped for regex if needed),

        to create a compiled regex to return.
        """
        name = [self.escape_name(self._name)]
        aliases = [self.escape_name(alias) for alias in self._aliases]
        pattern = r'|'.join(name + aliases)

        # Escape all whitespace with a single backslash.
        # This ensures that regexp in the prefix is treated as it was before
        # the actual regexp was changed to use the verbose syntax.
        prefix = re.sub(r"(\s)", r"\\\1", self._prefix)
        pattern = self.PATTERN_TEMPLATE.format(prefix=prefix, command=pattern)
        return re.compile(pattern, re.IGNORECASE | re.VERBOSE)


class NickCommand(NamedRuleMixin, Rule):
    """Nickname Command rule definition.

    A nickname command rule is a named rule with a twist: instead of a prefix,
    the rule is triggered when the line starts with a registered nickname (or
    one of its aliases). The command's name itself can have aliases too.

    Here is an example with the ``dummy`` nickname command:

    .. code-block:: irc

        <user> BotName: dummy
        <Bot> You just invoked the nick command 'dummy'
        <user> AliasBotName: dummy
        <Bot> You just invoked the nick command 'dummy'
        <user> BotName: dummy-alias
        <Bot> You just invoked the nick command 'dummy' (as 'dummy-alias')
        <user> AliasBotName: dummy-alias
        <Bot> You just invoked the nick command 'dummy' (as 'dummy-alias')

    Apart from that, it behaves exactly like a :class:`generic rule <Rule>`.
    """
    PATTERN_TEMPLATE = r"""
        ^
        $nickname[:,]? # Nickname.
        \s+({command}) # Command as group 1.
        (?:\s+         # Whitespace to end command.
        (              # Rest of the line as group 2.
        (?:(\S+))?     # Parameters 1-4 as groups 3-6.
        (?:\s+(\S+))?
        (?:\s+(\S+))?
        (?:\s+(\S+))?
        .*             # Accept anything after the parameters. Leave it up to
                       # the plugin to parse the line.
        ))?            # Group 1 must be None, if there are no parameters.
        $              # EoL, so there are no partial matches.
    """

    @classmethod
    def from_callable(cls, settings, handler):
        nick = settings.core.nick
        nick_aliases = tuple(settings.core.alias_nicks)
        commands = getattr(handler, 'nickname_commands', [])
        if not commands:
            raise RuntimeError('Invalid nick command callable: %s' % handler)

        name = commands[0]
        aliases = commands[1:]
        kwargs = cls.kwargs_from_callable(handler)
        kwargs.update({
            'nick': nick,
            'name': name,
            'nick_aliases': nick_aliases,
            'aliases': aliases,
            'handler': handler,
        })

        return cls(**kwargs)

    def __init__(self, nick, name, nick_aliases=None, aliases=None, **kwargs):
        super(NickCommand, self).__init__([], **kwargs)
        self._nick = nick
        self._name = name
        self._nick_aliases = (tuple(nick_aliases)
                              if nick_aliases is not None
                              else tuple())
        self._aliases = tuple(aliases) if aliases is not None else tuple()
        self._regexes = (self.get_rule_regex(),)

    def __str__(self):
        label = self.get_rule_label()
        plugin = self.get_plugin_name() or '(no-plugin)'
        aliases = '|'.join(self._aliases)
        nick = self._nick
        nick_aliases = '|'.join(self._nick_aliases)

        return '<NickCommand %s.%s [%s] (%s [%s])>' % (
            plugin, label, aliases, nick, nick_aliases)

    def get_usages(self):
        usages = []

        for usage in self._usages:
            text = usage.get('example')
            if not text:
                continue

            new_usage = {
                'text': text.replace('$nickname', self._nick),
                'result': usage.get('result', None),
                'is_pattern': bool(usage.get('is_pattern')),
                'is_admin': bool(usage.get('is_admin')),
                'is_owner': bool(usage.get('is_owner')),
                'is_private_message': bool(usage.get('is_private_message')),
            }
            usages.append(new_usage)

        return tuple(usages)

    def get_rule_regex(self):
        """Make the rule regex for this nick command.

        :return: a compiled regex for this nick command and its aliases

        The command regex factors in:

        * the nicks to react to,
        * the rule's name (escaped for regex),
        * all of its aliases (escaped for regex),

        to create a compiled regex to return.
        """
        name = [self.escape_name(self._name)]
        aliases = [self.escape_name(alias) for alias in self._aliases]
        pattern = r'|'.join(name + aliases)

        return _compile_pattern(
            self.PATTERN_TEMPLATE.format(command=pattern),
            self._nick,
            self._nick_aliases)


class ActionCommand(NamedRuleMixin, Rule):
    """Action Command rule definition.

    An action command rule is a named rule that can be triggered only when the
    trigger's intent is an ``ACTION``. Like the :class:`Command` rule, it
    allows command aliases.

    Here is an example with the ``dummy`` action command:

    .. code-block:: irc

        > user dummy
        <Bot> You just invoked the action command 'dummy'
        > user dummy-alias
        <Bot> You just invoked the action command 'dummy' (as 'dummy-alias')

    Apart from that, it behaves exactly like a :class:`generic rule <Rule>`.
    """
    INTENT_REGEX = re.compile(r'ACTION', re.IGNORECASE)
    PATTERN_TEMPLATE = r"""
        ({command})         # Command as group 1.
        (?:\s+              # Whitespace to end command.
        (                   # Rest of the line as group 2.
        (?:(\S+))?          # Parameters 1-4 as groups 3-6.
        (?:\s+(\S+))?
        (?:\s+(\S+))?
        (?:\s+(\S+))?
        .*                  # Accept anything after the parameters.
                            # Leave it up to the plugin to parse
                            # the line.
        ))?                 # Group 2 must be None, if there are no
                            # parameters.
        $                   # EoL, so there are no partial matches.
    """

    @classmethod
    def from_callable(cls, settings, handler):
        commands = getattr(handler, 'action_commands', [])
        if not commands:
            raise RuntimeError('Invalid action command callable: %s' % handler)

        name = commands[0]
        aliases = commands[1:]
        kwargs = cls.kwargs_from_callable(handler)
        kwargs.update({
            'name': name,
            'aliases': aliases,
            'handler': handler,
        })

        return cls(**kwargs)

    def __init__(self, name, aliases=None, **kwargs):
        super(ActionCommand, self).__init__([], **kwargs)
        self._name = name
        self._aliases = tuple(aliases) if aliases is not None else tuple()
        self._regexes = (self.get_rule_regex(),)

    def __str__(self):
        label = self.get_rule_label()
        plugin = self.get_plugin_name() or '(no-plugin)'
        aliases = '|'.join(self._aliases)

        return '<ActionCommand %s.%s [%s]>' % (plugin, label, aliases)

    def get_rule_regex(self):
        """Make the rule regex for this action command.

        :return: a compiled regex for this action command and its aliases

        The command regex factors in:

        * the rule's name (escaped for regex if needed),
        * all of its aliases (escaped for regex if needed),

        to create a compiled regex to return.
        """
        name = [self.escape_name(self._name)]
        aliases = [self.escape_name(alias) for alias in self._aliases]
        pattern = r'|'.join(name + aliases)
        pattern = self.PATTERN_TEMPLATE.format(command=pattern)
        return re.compile(pattern, re.IGNORECASE | re.VERBOSE)

    def match_intent(self, intent):
        """Tell if ``intent`` is an ``ACTION``.

        :param str intent: potential matching intent
        :return: ``True`` when ``intent`` matches ``ACTION``,
                 ``False`` otherwise
        :rtype: bool
        """
        return bool(intent and self.INTENT_REGEX.match(intent))


class FindRule(Rule):
    """Anonymous find rule definition.

    A find rule is like an anonymous rule with a twist: instead of matching
    only once per IRC line, a find rule will execute for each non-overlapping
    match for each of its regular expressions.

    For example, to match for each word starting with the letter ``h`` in a line,
    you can use the pattern ``h\\w+``:

    .. code-block:: irc

        <user> hello here
        <Bot> Found the word "hello"
        <Bot> Found the word "here"
        <user> sopelunker, how are you?
        <Bot> Found the word "how"

    .. seealso::

        This rule uses :func:`re.finditer`. To know more about how it works,
        see the official Python documentation.

    """
    REGEX_ATTRIBUTE = 'find_rules'
    LAZY_ATTRIBUTE = 'find_rules_lazy_loaders'

    def parse(self, text):
        for regex in self._regexes:
            for match in regex.finditer(text):
                yield match


class SearchRule(Rule):
    """Anonymous search rule definition.

    A search rule is like an anonymous rule with a twist: it will execute
    exactly once per regular expression that matches anywhere in a line, not
    just from the start.

    For example, to search if any word starts with the letter ``h`` in a line,
    you can use the pattern ``h\\w+``:

    .. code-block:: irc

        <user> hello here
        <Bot> Found the word "hello"
        <user> sopelunker, how are you?
        <Bot> Found the word "how"

    The match object it returns contains the first element that matches the
    expression in the line.

    .. seealso::

        This rule uses :func:`re.search`. To know more about how it works,
        see the official Python documentation.

    """
    REGEX_ATTRIBUTE = 'search_rules'
    LAZY_ATTRIBUTE = 'search_rules_lazy_loaders'

    def parse(self, text):
        for regex in self._regexes:
            match = regex.search(text)
            if match:
                yield match


class URLCallback(Rule):
    """URL callback rule definition.

    A URL callback rule (or simply "a URL rule") detects URLs in a trigger
    then it uses regular expressions to match at most once per URL per regular
    expression, i.e. you can trigger between 0 and the number of regex the URL
    callback has per URL in the IRC line.

    Here is an example with a URL rule with the pattern
    ``r'https://example\\.com/(.*)'``:

    .. code-block:: irc

        <user> https://example.com/test
        <Bot> You triggered a URL callback, with the "/test" path
        <user> and this URL is https://example.com/other can you get it?
        <Bot> You triggered a URL callback, with the "/other" path

    Like generic rules, URL callback rules are not triggered by any specific
    name and they don't have aliases.

    .. note::

        Unlike generic rules and commands, the :func:`~sopel.plugin.url`
        decorator expects its decorated function to have the bot and the
        trigger with a third parameter: the ``match`` parameter.

        To use this class with an existing URL callback handler, the
        :meth:`from_callable` classmethod **must** be used: it will wrap the
        handler to work as intended. In that case, the ``trigger`` and the
        ``match`` arguments will be the same when the rule executes.

        This behavior makes the ``match`` parameter obsolete, which will be
        removed in Sopel 9.

    """
    REGEX_ATTRIBUTE = 'url_regex'
    LAZY_ATTRIBUTE = 'url_lazy_loaders'

    @classmethod
    def from_callable(cls, settings, handler):
        execute_handler = handler
        regexes = cls.regex_from_callable(settings, handler)
        kwargs = cls.kwargs_from_callable(handler)

        # do we need to handle the match parameter?
        # old style URL callback: callable(bot, trigger, match)
        # new style: callable(bot, trigger)
        match_count = 3
        if inspect.ismethod(handler):
            # account for the 'self' parameter when the handler is a method
            match_count = 4

        argspec = inspect_getargspec(handler)

        if len(argspec.args) >= match_count:
            @functools.wraps(handler)
            def execute_handler(bot, trigger):
                return handler(bot, trigger, match=trigger)

        kwargs.update({
            'handler': execute_handler,
            'schemes': settings.core.auto_url_schemes,
        })

        return cls(regexes, **kwargs)

    @classmethod
    def from_callable_lazy(cls, settings, handler):
        """Instantiate a rule object from a handler with lazy-loaded regexes.

        :param settings: Sopel's settings
        :type settings: :class:`sopel.config.Config`
        :param callable handler: a function-based rule handler with a
                                 lazy-loader for the regexes
        :return: an instance of this class created from the ``handler``
        :rtype: :class:`AbstractRule`

        Similar to the :meth:`from_callable` classmethod, it requires a rule
        handlers decorated with :mod:`sopel.plugin`'s decorators.

        Unlike the :meth:`from_callable` classmethod, the regexes are not
        already attached to the handler: its loader functions will be used to
        get the rule's regexes. See the :func:`sopel.plugin.url_lazy` decorator
        for more information about the handler and the loaders' signatures.

        .. seealso::

            The handler can have more than one loader attached. In that case,
            these loaders are chained with :func:`sopel.tools.chain_loaders`.

        """
        regexes = cls.regex_from_callable_lazy(settings, handler)
        kwargs = cls.kwargs_from_callable(handler)
        kwargs.update({
            'handler': handler,
            'schemes': settings.core.auto_url_schemes,
        })

        return cls(regexes, **kwargs)

    def __init__(self,
                 regexes,
                 schemes=None,
                 **kwargs):
        super(URLCallback, self).__init__(regexes, **kwargs)
        # prevent mutability of registered schemes
        self._schemes = tuple(schemes or URL_DEFAULT_SCHEMES)

    def match(self, bot, pretrigger):
        """Match URL(s) in a pretrigger according to the rule.

        :param bot: Sopel instance
        :type bot: :class:`sopel.bot.Sopel`
        :param pretrigger: line to match
        :type pretrigger: :class:`sopel.trigger.PreTrigger`

        This method looks for :attr:`URLs in the IRC line
        <sopel.trigger.PreTrigger.urls>`, and for each it yields
        :ref:`match objects <match-objects>` using its regexes.

        .. seealso::

            To detect URLs, this method uses the
            :attr:`core.auto_url_schemes
            <sopel.config.core_section.CoreSection.auto_url_schemes>` option.

        """
        if not self.match_preconditions(bot, pretrigger):
            return

        # Parse only valid URLs with wanted schemes
        for url in pretrigger.urls:
            try:
                if urlparse(url).scheme not in self._schemes:
                    # skip URLs with unwanted scheme
                    continue
            except ValueError:
                # skip invalid URLs
                continue

            # TODO: convert to 'yield from' when dropping Python 2.7
            for result in self.parse(url):
                yield result

    def parse(self, text):
        for regex in self._regexes:
            result = regex.search(text)
            if result:
                yield result



# coding=utf-8
"""Tools to help safely do calculations from user input"""
from __future__ import absolute_import, division, print_function, unicode_literals

import ast
import numbers
import operator
import time

__all__ = ['eval_equation']


class ExpressionEvaluator:
    """A generic class for evaluating limited forms of Python expressions.

    :param dict bin_ops: optional; see notes
    :param dict unary_ops: optional; see notes

    Instances can pass ``binary_ops`` and ``unary_ops`` arguments with dicts of
    the form ``{ast.Node, function}``. When the :class:`ast.Node <ast.AST>` used
    as key is found, it will be evaluated using the given ``function``.
    """

    class Error(Exception):
        """Internal exception type for :class:`ExpressionEvaluator`\\s."""
        pass

    def __init__(self, bin_ops=None, unary_ops=None):
        self.binary_ops = bin_ops or {}
        self.unary_ops = unary_ops or {}

    def __call__(self, expression_str, timeout=5.0):
        """Evaluate a Python expression and return the result.

        :param str expression_str: the expression to evaluate
        :raise SyntaxError: if the given ``expression_str`` is not a valid
                            Python statement
        :raise ExpressionEvaluator.Error: if the instance of
            :class:`ExpressionEvaluator` does not have a handler for the
            :class:`ast.Node <ast.AST>`
        """
        ast_expression = ast.parse(expression_str, mode='eval')
        return self._eval_node(ast_expression.body, time.time() + timeout)

    def _eval_node(self, node, timeout):
        """Recursively evaluate the given :class:`ast.Node <ast.AST>`.

        :param node: the AST node to evaluate
        :type node: :class:`ast.AST`
        :param timeout: how long the expression is allowed to process before
                        timing out and failing
        :type timeout: int or float
        :raise ExpressionEvaluator.Error: if it can't handle the ``node``

        Uses :attr:`self.binary_ops` and :attr:`self.unary_ops` for the
        implementation.

        A subclass could overwrite this to handle more nodes, calling it only
        for nodes it does not implement itself.
        """
        if isinstance(node, ast.Num):
            return node.n

        elif (isinstance(node, ast.BinOp) and
                type(node.op) in self.binary_ops):
            left = self._eval_node(node.left, timeout)
            right = self._eval_node(node.right, timeout)
            if time.time() > timeout:
                raise ExpressionEvaluator.Error(
                    "Time for evaluating expression ran out.")
            return self.binary_ops[type(node.op)](left, right)

        elif (isinstance(node, ast.UnaryOp) and
                type(node.op) in self.unary_ops):
            operand = self._eval_node(node.operand, timeout)
            if time.time() > timeout:
                raise ExpressionEvaluator.Error(
                    "Time for evaluating expression ran out.")
            return self.unary_ops[type(node.op)](operand)

        raise ExpressionEvaluator.Error(
            "Ast.Node '%s' not implemented." % (type(node).__name__,))


def guarded_mul(left, right):
    """Multiply two values, guarding against overly large inputs.

    :param left: the left operand
    :type left: int or float
    :param right: the right operand
    :type right: int or float
    :raise ValueError: if the inputs are too large to handle safely
    """
    # Only handle ints because floats will overflow anyway.
    if not isinstance(left, numbers.Integral):
        pass
    elif not isinstance(right, numbers.Integral):
        pass
    elif left in (0, 1) or right in (0, 1):
        # Ignore trivial cases.
        pass
    elif left.bit_length() + right.bit_length() > 664386:
        # 664386 is the number of bits (10**100000)**2 has, which is instant on
        # my laptop, while (10**1000000)**2 has a noticeable delay. It could
        # certainly be improved.
        raise ValueError(
            "Value is too large to be handled in limited time and memory.")

    return operator.mul(left, right)


def pow_complexity(num, exp):
    """Estimate the worst case time :func:`pow` takes to calculate.

    :param num: base
    :type num: int or float
    :param exp: exponent
    :type exp: int or float

    This function is based on experimental data from the time it takes to
    calculate ``num**exp`` in 32-bit CPython 2.7.6 on an Intel Core i7-2670QM
    laptop running Windows.

    It tries to implement this surface: x=exp, y=num

    +----+------+------+------+------+-------+-------+-------+-------+-------+
    |    |  1e5 |  2e5 |  3e5 |  4e5 |  5e5  |  6e5  |  7e5  |  8e5  |  9e5  |
    +====+======+======+======+======+=======+=======+=======+=======+=======+
    | e1 | 0.03 | 0.09 | 0.16 | 0.25 |  0.35 |  0.46 |  0.60 |  0.73 |  0.88 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e2 | 0.08 | 0.24 | 0.46 | 0.73 |  1.03 |  1.40 |  1.80 |  2.21 |  2.63 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e3 | 0.15 | 0.46 | 0.87 | 1.39 |  1.99 |  2.63 |  3.35 |  4.18 |  5.15 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e4 | 0.24 | 0.73 | 1.39 | 2.20 |  3.11 |  4.18 |  5.39 |  6.59 |  7.88 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e5 | 0.34 | 1.03 | 2.00 | 3.12 |  4.48 |  5.97 |  7.56 |  9.37 | 11.34 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e6 | 0.46 | 1.39 | 2.62 | 4.16 |  5.97 |  7.86 | 10.09 | 12.56 | 15.39 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e7 | 0.60 | 1.79 | 3.34 | 5.39 |  7.60 | 10.16 | 13.00 | 16.23 | 19.44 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e8 | 0.73 | 2.20 | 4.18 | 6.60 |  9.37 | 12.60 | 16.26 | 19.83 | 23.70 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+
    | e9 | 0.87 | 2.62 | 5.15 | 7.93 | 11.34 | 15.44 | 19.40 | 23.66 | 28.58 |
    +----+------+------+------+------+-------+-------+-------+-------+-------+

    For powers of 2 it tries to implement this surface:

    +---+------+------+------+------+------+------+------+------+------+
    |   |  1e7 |  2e7 |  3e7 |  4e7 |  5e7 |  6e7 |  7e7 |  8e7 |  9e7 |
    +===+======+======+======+======+======+======+======+======+======+
    | 1 | 0.00 | 0.00 | 0.00 | 0.00 | 0.00 | 0.00 | 0.00 | 0.00 | 0.00 |
    +---+------+------+------+------+------+------+------+------+------+
    | 2 | 0.21 | 0.44 | 0.71 | 0.92 | 1.20 | 1.49 | 1.66 | 1.95 | 2.23 |
    +---+------+------+------+------+------+------+------+------+------+
    | 4 | 0.43 | 0.91 | 1.49 | 1.96 | 2.50 | 3.13 | 3.54 | 4.10 | 4.77 |
    +---+------+------+------+------+------+------+------+------+------+
    | 8 | 0.70 | 1.50 | 2.24 | 3.16 | 3.83 | 4.66 | 5.58 | 6.56 | 7.67 |
    +---+------+------+------+------+------+------+------+------+------+

    The function numbers were selected by starting with the theoretical
    complexity of ``exp * log2(num)**2`` and fiddling with the exponents
    until it more or less matched with the table.

    Because this function is based on a limited set of data it might not give
    accurate results outside these boundaries. The results derived from large
    ``num`` and ``exp`` were quite accurate for small ``num`` and very large
    ``exp`` though, except when ``num`` was a power of 2.
    """
    if num in (0, 1) or exp in (0, 1):
        return 0
    elif (num & (num - 1)) == 0:
        # For powers of 2 the scaling is a bit different.
        return exp ** 1.092 * num.bit_length() ** 1.65 / 623212911.121
    else:
        return exp ** 1.590 * num.bit_length() ** 1.73 / 36864057619.3


def guarded_pow(num, exp):
    """Raise a number to a power, guarding against overly large inputs.

    :param num: base
    :type num: int or float
    :param exp: exponent
    :type exp: int or float
    :raise ValueError: if the inputs are too large to handle safely
    """
    # Only handle ints because floats will overflow anyway.
    if not isinstance(num, numbers.Integral):
        pass
    elif not isinstance(exp, numbers.Integral):
        pass
    elif pow_complexity(num, exp) < 0.5:
        # Value 0.5 is arbitrary and based on an estimated runtime of 0.5s
        # on a fairly decent laptop processor.
        pass
    else:
        raise ValueError("Pow expression too complex to calculate.")

    return operator.pow(num, exp)


class EquationEvaluator(ExpressionEvaluator):
    __bin_ops = {
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: guarded_mul,
        ast.Div: operator.truediv,
        ast.Pow: guarded_pow,
        ast.Mod: operator.mod,
        ast.FloorDiv: operator.floordiv,
        ast.BitXor: guarded_pow
    }
    __unary_ops = {
        ast.USub: operator.neg,
        ast.UAdd: operator.pos,
    }

    def __init__(self):
        ExpressionEvaluator.__init__(
            self,
            bin_ops=self.__bin_ops,
            unary_ops=self.__unary_ops
        )

    def __call__(self, expression_str):
        result = ExpressionEvaluator.__call__(self, expression_str)

        # This wrapper is here so additional sanity checks could be done
        # on the result of the eval, but currently none are done.

        return result


eval_equation = EquationEvaluator()
"""Evaluates a Python equation expression and returns the result.

:param str equation: the equation to evaluate
:param timeout: optional timeout value
:type timeout: int or float

Supports addition (+), subtraction (-), multiplication (*), division (/),
power (**) and modulo (%).
"""



# coding=utf-8
""":mod:`sopel.irc` is the core IRC module for Sopel.

This sub-package contains everything that is related to the IRC protocol
(connection, commands, abstract client, etc.) and that can be used to implement
the Sopel bot.

In particular, it defines the interface for the
:class:`IRC backend<sopel.irc.abstract_backends.AbstractIRCBackend>`, and the
interface for the :class:`bot itself<sopel.irc.AbstractBot>`. This is all
internal code that isn't supposed to be used directly by a plugin developer,
who should worry about :class:`sopel.bot.Sopel` only.

.. important::

    When working on core IRC protocol related features, consult protocol
    documentation at https://www.irchelp.org/protocol/rfc/

"""
# Copyright 2008, Sean B. Palmer, inamidst.com
# Copyright 2012, Elsie Powell, http://embolalia.com
# Copyright © 2012, Elad Alfassa <elad@fedoraproject.org>
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime
import logging
import os
import sys
import threading
import time

try:
    import ssl
    if not hasattr(ssl, 'match_hostname'):
        # Attempt to import ssl_match_hostname from python-backports
        # TODO: Remove when dropping Python 2 support
        import backports.ssl_match_hostname
        ssl.match_hostname = backports.ssl_match_hostname.match_hostname
        ssl.CertificateError = backports.ssl_match_hostname.CertificateError
    has_ssl = True
except ImportError:
    # no SSL support
    has_ssl = False

from sopel import tools, trigger
from .backends import AsynchatBackend, SSLAsynchatBackend
from .isupport import ISupport
from .utils import CapReq, safe

if sys.version_info.major >= 3:
    unicode = str

__all__ = ['abstract_backends', 'backends', 'utils']

LOGGER = logging.getLogger(__name__)


class AbstractBot(object):
    """Abstract definition of Sopel's interface."""
    def __init__(self, settings):
        # private properties: access as read-only properties
        self._nick = tools.Identifier(settings.core.nick)
        self._user = settings.core.user
        self._name = settings.core.name
        self._isupport = ISupport()
        self._myinfo = None

        self.backend = None
        """IRC Connection Backend."""
        self.connection_registered = False
        """Is the IRC Connection registered yet?"""
        self.settings = settings
        """Bot settings."""
        self.enabled_capabilities = set()
        """A set containing the IRCv3 capabilities that the bot has enabled."""
        self._cap_reqs = dict()
        """A dictionary of capability names to a list of requests."""

        # internal machinery
        self.sending = threading.RLock()
        self.last_error_timestamp = None
        self.error_count = 0
        self.stack = {}
        self.hasquit = False
        self.last_raw_line = ''  # last raw line received

    @property
    def nick(self):
        """Sopel's current nick.

        Changing this while Sopel is running is unsupported.
        """
        return self._nick

    @property
    def user(self):
        """Sopel's user/ident."""
        return self._user

    @property
    def name(self):
        """Sopel's "real name", as used for WHOIS responses."""
        return self._name

    @property
    def config(self):
        """The :class:`sopel.config.Config` for the current Sopel instance."""
        # TODO: Deprecate config, replaced by settings
        return self.settings

    @property
    def isupport(self):
        """Features advertised by the server.

        :type: :class:`~.isupport.ISupport` instance
        """
        return self._isupport

    @property
    def myinfo(self):
        """Server/network information.

        :type: :class:`~.utils.MyInfo` instance

        .. versionadded:: 7.0
        """
        if self._myinfo is None:
            raise AttributeError('myinfo')
        return self._myinfo

    # Connection

    def get_irc_backend(self):
        """Set up the IRC backend based on the bot's settings.

        :return: the initialized IRC backend object
        :rtype: an object implementing the interface of
                :class:`~sopel.irc.abstract_backends.AbstractIRCBackend`
        """
        timeout = int(self.settings.core.timeout)
        ping_interval = int(self.settings.core.timeout_ping_interval)
        backend_class = AsynchatBackend
        backend_args = [self]
        backend_kwargs = {
            'server_timeout': timeout,
            'ping_interval': ping_interval,
        }

        if self.settings.core.use_ssl:
            if has_ssl:
                backend_class = SSLAsynchatBackend
                backend_kwargs.update({
                    'verify_ssl': self.settings.core.verify_ssl,
                    'ca_certs': self.settings.core.ca_certs,
                })
            else:
                LOGGER.warning(
                    'SSL is not available on your system; '
                    'attempting connection without it')

        return backend_class(*backend_args, **backend_kwargs)

    def run(self, host, port=6667):
        """Connect to IRC server and run the bot forever.

        :param str host: the IRC server hostname
        :param int port: the IRC server port
        """
        source_address = ((self.settings.core.bind_host, 0)
                          if self.settings.core.bind_host else None)

        self.backend = self.get_irc_backend()
        self.backend.initiate_connect(host, port, source_address)
        try:
            self.backend.run_forever()
        except KeyboardInterrupt:
            # raised only when the bot is not connected
            LOGGER.warning('Keyboard Interrupt')
            raise

    def on_connect(self):
        """Handle successful establishment of IRC connection."""
        LOGGER.info('Connected, initiating setup sequence')

        # Request list of server capabilities. IRCv3 servers will respond with
        # CAP * LS (which we handle in coretasks). v2 servers will respond with
        # 421 Unknown command, which we'll ignore
        LOGGER.debug('Sending CAP request')
        self.backend.send_command('CAP', 'LS', '302')

        # authenticate account if needed
        if self.settings.core.auth_method == 'server':
            LOGGER.debug('Sending server auth')
            self.backend.send_pass(self.settings.core.auth_password)
        elif self.settings.core.server_auth_method == 'server':
            LOGGER.debug('Sending server auth')
            self.backend.send_pass(self.settings.core.server_auth_password)

        LOGGER.debug('Sending nick "%s"', self.nick)
        self.backend.send_nick(self.nick)
        LOGGER.debug('Sending user "%s" (name: "%s")', self.user, self.name)
        self.backend.send_user(self.user, '0', '*', self.name)

    def on_message(self, message):
        """Handle an incoming IRC message.

        :param str message: the received raw IRC message
        """
        self.last_raw_line = message

        pretrigger = trigger.PreTrigger(
            self.nick,
            message,
            url_schemes=self.settings.core.auto_url_schemes,
        )
        if all(cap not in self.enabled_capabilities for cap in ['account-tag', 'extended-join']):
            pretrigger.tags.pop('account', None)

        if pretrigger.event == 'PING':
            self.backend.send_pong(pretrigger.args[-1])
        elif pretrigger.event == 'ERROR':
            LOGGER.error("ERROR received from server: %s", pretrigger.args[-1])
            self.backend.on_irc_error(pretrigger)

        self.dispatch(pretrigger)

    def on_message_sent(self, raw):
        """Handle any message sent through the connection.

        :param str raw: raw text message sent through the connection

        When a message is sent through the IRC connection, the bot will log
        the raw message. If necessary, it will also simulate the
        `echo-message`_ feature of IRCv3.

        .. _echo-message: https://ircv3.net/irc/#echo-message
        """
        # Log raw message
        self.log_raw(raw, '>>')

        # Simulate echo-message
        no_echo = 'echo-message' not in self.enabled_capabilities
        echoed = ['PRIVMSG', 'NOTICE']
        if no_echo and any(raw.upper().startswith(cmd) for cmd in echoed):
            # Use the hostmask we think the IRC server is using for us,
            # or something reasonable if that's not available
            host = 'localhost'
            if self.settings.core.bind_host:
                host = self.settings.core.bind_host
            else:
                try:
                    host = self.hostmask
                except KeyError:
                    pass  # we tried, and that's good enough

            pretrigger = trigger.PreTrigger(
                self.nick,
                ":{0}!{1}@{2} {3}".format(self.nick, self.user, host, raw),
                url_schemes=self.settings.core.auto_url_schemes,
            )
            self.dispatch(pretrigger)

    def on_error(self):
        """Handle any uncaptured error in the bot itself."""
        LOGGER.error('Fatal error in core, please review exceptions log.')

        err_log = logging.getLogger('sopel.exceptions')
        err_log.error(
            'Fatal error in core, handle_error() was called.\n'
            'Buffer:\n%s\n'
            'Last Line:\n%s',
            self.backend.buffer,  # TODO: refactor without self.backend
            self.last_raw_line,
        )
        err_log.exception('Fatal error traceback')
        err_log.error('----------------------------------------')

        if self.error_count > 10:
            # quit if too many errors
            dt = datetime.utcnow() - self.last_error_timestamp
            dt_seconds = dt.total_seconds()
            if dt_seconds < 5:
                LOGGER.error('Too many errors, can\'t continue')
                os._exit(1)
            # remove 1 error per full 5s that passed since last error
            self.error_count = max(0, self.error_count - dt_seconds // 5)

        self.last_error_timestamp = datetime.utcnow()
        self.error_count = self.error_count + 1

    def change_current_nick(self, new_nick):
        """Change the current nick without configuration modification.

        :param str new_nick: new nick to be used by the bot
        """
        self._nick = tools.Identifier(new_nick)
        LOGGER.debug('Sending nick "%s"', self.nick)
        self.backend.send_nick(self.nick)

    def on_close(self):
        """Call shutdown methods."""
        self._shutdown()

    def _shutdown(self):
        """Handle shutdown tasks.

        Must be overridden by subclasses to do anything useful.
        """
        pass

    # Features

    def dispatch(self, pretrigger):
        """Handle running the appropriate callables for an incoming message.

        :param pretrigger: Sopel PreTrigger object
        :type pretrigger: :class:`sopel.trigger.PreTrigger`
        :raise NotImplementedError: if the subclass does not implement this
                                    required method

        .. important::
            This method **MUST** be implemented by concrete subclasses.
        """
        raise NotImplementedError

    def log_raw(self, line, prefix):
        """Log raw line to the raw log.

        :param str line: the raw line
        :param str prefix: additional information to prepend to the log line
        """
        if not self.settings.core.log_raw:
            return
        logger = logging.getLogger('sopel.raw')
        logger.info('\t'.join([prefix, line.strip()]))

    def cap_req(self, plugin_name, capability, arg=None, failure_callback=None,
                success_callback=None):
        """Tell Sopel to request a capability when it starts.

        :param str plugin_name: the plugin requesting the capability
        :param str capability: the capability requested, optionally prefixed
                               with ``-`` or ``=``
        :param str arg: arguments for the capability request
        :param failure_callback: a function that will be called if the
                                 capability request fails
        :type failure_callback: :term:`function`
        :param success_callback: a function that will be called if the
                                 capability is successfully requested
        :type success_callback: :term:`function`

        By prefixing the capability with ``-``, it will be ensured that the
        capability is not enabled. Similarly, by prefixing the capability with
        ``=``, it will be ensured that the capability is enabled. Requiring and
        disabling is "first come, first served"; if one plugin requires a
        capability, and another prohibits it, this function will raise an
        exception in whichever plugin loads second. An exception will also be
        raised if the plugin is being loaded after the bot has already started,
        and the request would change the set of enabled capabilities.

        If the capability is not prefixed, and no other plugin prohibits it, it
        will be requested. Otherwise, it will not be requested. Since
        capability requests that are not mandatory may be rejected by the
        server, as well as by other plugins, a plugin which makes such a
        request should account for that possibility.

        The actual capability request to the server is handled after the
        completion of this function. In the event that the server denies a
        request, the ``failure_callback`` function will be called, if provided.
        The arguments will be a :class:`~sopel.bot.Sopel` object, and the
        capability which was rejected. This can be used to disable callables
        which rely on the capability. It will be be called either if the server
        NAKs the request, or if the server enabled it and later DELs it.

        The ``success_callback`` function will be called upon acknowledgment
        of the capability from the server, whether during the initial
        capability negotiation, or later.

        If ``arg`` is given, and does not exactly match what the server
        provides or what other plugins have requested for that capability, it is
        considered a conflict.
        """
        # TODO raise better exceptions
        cap = capability[1:]
        prefix = capability[0]

        entry = self._cap_reqs.get(cap, [])
        if any((ent.arg != arg for ent in entry)):
            raise Exception('Capability conflict')

        if prefix == '-':
            if self.connection_registered and cap in self.enabled_capabilities:
                raise Exception('Can not change capabilities after server '
                                'connection has been completed.')
            if any((ent.prefix != '-' for ent in entry)):
                raise Exception('Capability conflict')
            entry.append(CapReq(prefix, plugin_name, failure_callback, arg,
                                success_callback))
            self._cap_reqs[cap] = entry
        else:
            if prefix != '=':
                cap = capability
                prefix = ''
            if self.connection_registered and (cap not in
                                               self.enabled_capabilities):
                raise Exception('Can not change capabilities after server '
                                'connection has been completed.')
            # Non-mandatory will callback at the same time as if the server
            # rejected it.
            if any((ent.prefix == '-' for ent in entry)) and prefix == '=':
                raise Exception('Capability conflict')
            entry.append(CapReq(prefix, plugin_name, failure_callback, arg,
                                success_callback))
            self._cap_reqs[cap] = entry

    def write(self, args, text=None):
        """Send a command to the server.

        :param args: an iterable of strings, which will be joined by spaces
        :type args: :term:`iterable`
        :param str text: a string that will be prepended with a ``:`` and added
                         to the end of the command

        ``args`` is an iterable of strings, which are joined by spaces.
        ``text`` is treated as though it were the final item in ``args``, but
        is preceded by a ``:``. This is a special case which means that
        ``text``, unlike the items in ``args``, may contain spaces (though this
        constraint is not checked by ``write``).

        In other words, both ``sopel.write(('PRIVMSG',), 'Hello, world!')``
        and ``sopel.write(('PRIVMSG', ':Hello, world!'))`` will send
        ``PRIVMSG :Hello, world!`` to the server.

        Newlines and carriage returns (``'\\n'`` and ``'\\r'``) are removed
        before sending. Additionally, if the message (after joining) is longer
        than than 510 characters, any remaining characters will not be sent.

        .. seealso::

            The connection backend is responsible for formatting and sending
            the message through the IRC connection. See the
            :meth:`sopel.irc.abstract_backends.AbstractIRCBackend.send_command`
            method for more information.

        """
        args = [safe(arg) for arg in args]
        self.backend.send_command(*args, text=text)

    # IRC Commands

    def action(self, text, dest):
        """Send a CTCP ACTION PRIVMSG to a user or channel.

        :param str text: the text to send in the CTCP ACTION
        :param str dest: the destination of the CTCP ACTION

        The same loop detection and length restrictions apply as with
        :func:`say`, though automatic message splitting is not available.
        """
        self.say('\001ACTION {}\001'.format(text), dest)

    def join(self, channel, password=None):
        """Join a ``channel``.

        :param str channel: the channel to join
        :param str password: an optional channel password

        If ``channel`` contains a space, and no ``password`` is given, the
        space is assumed to split the argument into the channel to join and its
        password. ``channel`` should not contain a space if ``password``
        is given.
        """
        self.backend.send_join(channel, password=password)

    def kick(self, nick, channel, text=None):
        """Kick a ``nick`` from a ``channel``.

        :param str nick: nick to kick out of the ``channel``
        :param str channel: channel to kick ``nick`` from
        :param str text: optional text for the kick

        The bot must be operator in the specified channel for this to work.

        .. versionadded:: 7.0
        """
        self.backend.send_kick(channel, nick, reason=text)

    def notice(self, text, dest):
        """Send an IRC NOTICE to a user or channel (``dest``).

        :param str text: the text to send in the NOTICE
        :param str dest: the destination of the NOTICE
        """
        self.backend.send_notice(dest, text)

    def part(self, channel, msg=None):
        """Leave a channel.

        :param str channel: the channel to leave
        :param str msg: the message to display when leaving a channel
        """
        self.backend.send_part(channel, reason=msg)

    def quit(self, message):
        """Disconnect from IRC and close the bot."""
        self.backend.send_quit(reason=message)
        self.hasquit = True
        # Wait for acknowledgment from the server. Per RFC 2812 it should be
        # an ERROR message, but many servers just close the connection.
        # Either way is fine by us. Closing the connection now would mean that
        # stuff in the buffers that has not yet been processed would never be
        # processed. It would also release the main thread, which is
        # problematic because whomever called quit might still want to do
        # something before the main thread quits.

    def reply(self, text, dest, reply_to, notice=False):
        """Send a PRIVMSG to a user or channel, prepended with ``reply_to``.

        :param str text: the text of the reply
        :param str dest: the destination of the reply
        :param str reply_to: the nickname that the reply will be prepended with
        :param bool notice: whether to send the reply as a NOTICE or not,
                            defaults to ``False``

        If ``notice`` is ``True``, send a NOTICE rather than a PRIVMSG.

        The same loop detection and length restrictions apply as with
        :meth:`say`, though automatic message splitting is not available.
        """
        text = '%s: %s' % (reply_to, text)
        if notice:
            self.notice(text, dest)
        else:
            self.say(text, dest)

    def say(self, text, recipient, max_messages=1, truncation='', trailing=''):
        """Send a PRIVMSG to a user or channel.

        :param str text: the text to send
        :param str recipient: the message recipient
        :param int max_messages: split ``text`` into at most this many messages
                                 if it is too long to fit in one (optional)
        :param str truncation: string to append if ``text`` is too long to fit
                               in a single message, or into the last message if
                               ``max_messages`` is greater than 1 (optional)
        :param str trailing: string to append after ``text`` and (if used)
                             ``truncation`` (optional)

        By default, this will attempt to send the entire ``text`` in one
        message. If the text is too long for the server, it may be truncated.

        If ``max_messages`` is given, the ``text`` will be split into at most
        that many messages. The split is made at the last space character
        before the "safe length" (which is calculated based on the bot's
        nickname and hostmask), or exactly at the "safe length" if no such
        space character exists.

        If the ``text`` is too long to fit into the specified number of messages
        using the above splitting, the final message will contain the entire
        remainder, which may be truncated by the server. You can specify
        ``truncation`` to tell Sopel how it should indicate that the remaining
        ``text`` was cut off. Note that the ``truncation`` parameter must
        include leading whitespace if you desire any between it and the
        truncated text.

        The ``trailing`` parameter is *always* appended to ``text``, after the
        point where ``truncation`` would be inserted if necessary. It's useful
        for making sure e.g. a link is always included, even if the summary your
        plugin fetches is too long to fit.

        Here are some examples of how the ``truncation`` and ``trailing``
        parameters work, using an artificially low maximum line length::

            # bot.say() outputs <text> + <truncation?> + <trailing>
            #                   always     if needed       always

            bot.say(
                '"This is a short quote.',
                truncation=' […]',
                trailing='"')
            # Sopel says: "This is a short quote."

            bot.say(
                '"This quote is very long and will not fit on a line.',
                truncation=' […]',
                trailing='"')
            # Sopel says: "This quote is very long […]"

            bot.say(
                # note the " included at the end this time
                '"This quote is very long and will not fit on a line."',
                truncation=' […]')
            # Sopel says: "This quote is very long […]
            # The ending " goes missing

        .. versionadded:: 7.1

            The ``truncation`` and ``trailing`` parameters.

        """
        excess = ''
        if not isinstance(text, unicode):
            # Make sure we are dealing with a Unicode string
            text = text.decode('utf-8')

        if max_messages > 1 or truncation or trailing:
            # Handle message splitting/truncation only if needed
            try:
                hostmask_length = len(self.hostmask)
            except KeyError:
                # calculate maximum possible length, given current nick/username
                hostmask_length = (
                    len(self.nick)  # own nick length
                    + 1  # (! separator)
                    + 1  # (for the optional ~ in user)
                    + min(  # own ident length, capped to ISUPPORT or RFC maximum
                        len(self.user),
                        getattr(self.isupport, 'USERLEN', 9))
                    + 1  # (@ separator)
                    + 63  # <hostname> has a maximum length of 63 characters.
                )
            safe_length = (
                512  # maximum IRC line length in bytes, per RFC
                - 1  # leading colon
                - hostmask_length  # calculated/maximum length of own hostmask prefix
                - 1  # space between prefix & command
                - 7  # PRIVMSG command
                - 1  # space before recipient
                - len(recipient.encode('utf-8'))  # target channel/nick (can contain Unicode)
                - 2  # space after recipient, colon before text
                - 2  # trailing CRLF
            )

            if max_messages == 1 and trailing:
                safe_length -= len(trailing.encode('utf-8'))
            text, excess = tools.get_sendable_message(text, safe_length)

        if max_messages == 1:
            if excess and truncation:
                # only append `truncation` if this is the last message AND it's still too long
                safe_length -= len(truncation.encode('utf-8'))
                text, excess = tools.get_sendable_message(text, safe_length)
                text += truncation

            # ALWAYS append `trailing`;
            # its length is included when determining if truncation happened above
            text += trailing

        flood_max_wait = self.settings.core.flood_max_wait
        flood_burst_lines = self.settings.core.flood_burst_lines
        flood_refill_rate = self.settings.core.flood_refill_rate
        flood_empty_wait = self.settings.core.flood_empty_wait

        flood_text_length = self.settings.core.flood_text_length
        flood_penalty_ratio = self.settings.core.flood_penalty_ratio

        with self.sending:
            recipient_id = tools.Identifier(recipient)
            recipient_stack = self.stack.setdefault(recipient_id, {
                'messages': [],
                'flood_left': flood_burst_lines,
            })

            if recipient_stack['messages']:
                elapsed = time.time() - recipient_stack['messages'][-1][0]
            else:
                # Default to a high enough value that we won't care.
                # Five minutes should be enough not to matter anywhere below.
                elapsed = 300

            # If flood bucket is empty, refill the appropriate number of lines
            # based on how long it's been since our last message to recipient
            if not recipient_stack['flood_left']:
                recipient_stack['flood_left'] = min(
                    flood_burst_lines,
                    int(elapsed) * flood_refill_rate)

            # If it's too soon to send another message, wait
            if not recipient_stack['flood_left']:
                penalty = 0

                if flood_penalty_ratio > 0:
                    penalty_ratio = flood_text_length * flood_penalty_ratio
                    text_length_overflow = float(
                        max(0, len(text) - flood_text_length))
                    penalty = text_length_overflow / penalty_ratio

                # Maximum wait time is 2 sec by default
                initial_wait_time = flood_empty_wait + penalty
                wait = min(initial_wait_time, flood_max_wait)
                if elapsed < wait:
                    sleep_time = wait - elapsed
                    LOGGER.debug(
                        'Flood protection wait time: %.3fs; '
                        'elapsed time: %.3fs; '
                        'initial wait time (limited to %.3fs): %.3fs '
                        '(including %.3fs of penalty).',
                        sleep_time,
                        elapsed,
                        flood_max_wait,
                        initial_wait_time,
                        penalty,
                    )
                    time.sleep(sleep_time)

            # Loop detection
            messages = [m[1] for m in recipient_stack['messages'][-8:]]

            # If what we're about to send repeated at least 5 times in the last
            # two minutes, replace it with '...'
            if messages.count(text) >= 5 and elapsed < 120:
                text = '...'
                if messages.count('...') >= 3:
                    # If we've already said '...' 3 times, discard message
                    return

            self.backend.send_privmsg(recipient, text)
            recipient_stack['flood_left'] = max(0, recipient_stack['flood_left'] - 1)
            recipient_stack['messages'].append((time.time(), safe(text)))
            recipient_stack['messages'] = recipient_stack['messages'][-10:]

        # Now that we've sent the first part, we need to send the rest if
        # requested. Doing so recursively seems simpler than iteratively.
        if max_messages > 1 and excess:
            self.say(excess, recipient, max_messages - 1, truncation, trailing)



# coding=utf-8
"""
help.py - Sopel Help Plugin
Copyright 2008, Sean B. Palmer, inamidst.com
Copyright © 2013, Elad Alfassa, <elad@fedoraproject.org>
Copyright © 2018, Adam Erdman, pandorah.org
Copyright © 2019, Tomasz Kurcz, github.com/uint
Copyright © 2019, dgw, technobabbl.es
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import collections
import logging
import re
import socket
import textwrap

import requests

from sopel import plugin, tools
from sopel.config import types


SETTING_CACHE_NAMESPACE = 'help-setting-cache'  # Set top-level memory key name
LOGGER = logging.getLogger(__name__)

# Settings that should require the help listing to be regenerated, or
# re-POSTed to paste, if they are changed during runtime.
# Keys are plugin names, and values are lists of setting names
# specific to that plugin.
TRACKED_SETTINGS = {
    'help': [
        'output',
        'show_server_host',
    ]
}


class PostingException(Exception):
    """Custom exception type for errors posting help to the chosen pastebin."""
    pass


# Pastebin handlers


def _requests_post_catch_errors(*args, **kwargs):
    try:
        response = requests.post(*args, **kwargs)
        response.raise_for_status()
    except (
            requests.exceptions.Timeout,
            requests.exceptions.TooManyRedirects,
            requests.exceptions.RequestException,
            requests.exceptions.HTTPError
    ):
        # We re-raise all expected exception types to a generic "posting error"
        # that's easy for callers to expect, and then we pass the original
        # exception through to provide some debugging info
        LOGGER.exception('Error during POST request')
        raise PostingException('Could not communicate with remote service')

    # remaining handling (e.g. errors inside the response) is left to the caller
    return response


def post_to_clbin(msg):
    try:
        result = _requests_post_catch_errors('https://clbin.com/', data={'clbin': msg})
    except PostingException:
        raise

    result = result.text
    if re.match(r'https?://clbin\.com/', result):
        # find/replace just in case the site tries to be sneaky and save on SSL overhead,
        # though it will probably send us an HTTPS link without any tricks.
        return result.replace('http://', 'https://', 1)
    else:
        LOGGER.error("Invalid result %s", result)
        raise PostingException('clbin result did not contain expected URL base.')


def post_to_0x0(msg):
    try:
        result = _requests_post_catch_errors('https://0x0.st', files={'file': msg})
    except PostingException:
        raise

    result = result.text
    if re.match(r'https?://0x0\.st/', result):
        # find/replace just in case the site tries to be sneaky and save on SSL overhead,
        # though it will probably send us an HTTPS link without any tricks.
        return result.replace('http://', 'https://', 1)
    else:
        LOGGER.error('Invalid result %s', result)
        raise PostingException('0x0.st result did not contain expected URL base.')


def post_to_hastebin(msg):
    try:
        result = _requests_post_catch_errors('https://hastebin.com/documents', data=msg)
    except PostingException:
        raise

    try:
        result = result.json()
    except ValueError:
        LOGGER.error("Invalid Hastebin response %s", result)
        raise PostingException('Could not parse response from Hastebin!')

    if 'key' not in result:
        LOGGER.error("Invalid result %s", result)
        raise PostingException('Hastebin result did not contain expected URL base.')
    return "https://hastebin.com/" + result['key']


def post_to_termbin(msg):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(10)  # the bot may NOT wait forever for a response; that would be bad
    try:
        sock.connect(('termbin.com', 9999))
        sock.sendall(msg)
        sock.shutdown(socket.SHUT_WR)
        response = ""
        while 1:
            data = sock.recv(1024)
            if data == "":
                break
            response += data
        sock.close()
    except socket.error:
        LOGGER.exception('Error during communication with termbin')
        raise PostingException('Error uploading to termbin')

    # find/replace just in case the site tries to be sneaky and save on SSL overhead,
    # though it will probably send us an HTTPS link without any tricks.
    return response.strip('\x00\n').replace('http://', 'https://', 1)


def post_to_ubuntu(msg):
    data = {
        'poster': 'sopel',
        'syntax': 'text',
        'expiration': '',
        'content': msg,
    }
    result = _requests_post_catch_errors(
        'https://pastebin.ubuntu.com/', data=data)

    if not re.match(r'https://pastebin\.ubuntu\.com/p/[^/]+/', result.url):
        LOGGER.error("Invalid Ubuntu pastebin response url %s", result.url)
        raise PostingException(
            'Invalid response from Ubuntu pastebin: %s' % result.url)

    return result.url


PASTEBIN_PROVIDERS = {
    'clbin': post_to_clbin,
    '0x0': post_to_0x0,
    'hastebin': post_to_hastebin,
    'termbin': post_to_termbin,
    'ubuntu': post_to_ubuntu,
}
REPLY_METHODS = [
    'channel',
    'query',
    'notice',
]


class HelpSection(types.StaticSection):
    """Configuration section for this plugin."""
    output = types.ChoiceAttribute('output',
                                   list(PASTEBIN_PROVIDERS),
                                   default='clbin')
    """The pastebin provider to use for help output."""
    reply_method = types.ChoiceAttribute('reply_method',
                                         REPLY_METHODS,
                                         default='channel')
    """Where/how to reply to help commands (public/private)."""
    show_server_host = types.BooleanAttribute('show_server_host',
                                              default=True)
    """Show the IRC server's hostname/IP in the first line of the help listing?"""


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | output | clbin | The pastebin provider to use for help output |
    | reply\\_method | channel | How/where help output should be sent |
    | show\\_server\\_host | True | Whether to show the IRC server's hostname/IP at the top of command listings |
    """
    config.define_section('help', HelpSection)
    provider_list = ', '.join(PASTEBIN_PROVIDERS)
    reply_method_list = ', '.join(REPLY_METHODS)
    config.help.configure_setting(
        'output',
        'Pick a pastebin provider: {}: '.format(provider_list)
    )
    config.help.configure_setting(
        'reply_method',
        'How/where should help command replies be sent: {}? '.format(reply_method_list)
    )
    config.help.configure_setting(
        'show_server_host',
        'Should the help command show the IRC server\'s hostname/IP in the listing?'
    )


def setup(bot):
    bot.config.define_section('help', HelpSection)

    # Initialize memory
    if SETTING_CACHE_NAMESPACE not in bot.memory:
        bot.memory[SETTING_CACHE_NAMESPACE] = tools.SopelMemory()

    # Initialize settings cache
    for section in TRACKED_SETTINGS:
        if section not in bot.memory[SETTING_CACHE_NAMESPACE]:
            bot.memory[SETTING_CACHE_NAMESPACE][section] = tools.SopelMemory()

    update_cache(bot)  # Populate cache

    bot.config.define_section('help', HelpSection)


def update_cache(bot):
    for section, setting_names_list in TRACKED_SETTINGS.items():
        for setting_name in setting_names_list:
            bot.memory[SETTING_CACHE_NAMESPACE][section][setting_name] = getattr(getattr(bot.config, section), setting_name)


def is_cache_valid(bot):
    for section, setting_names_list in TRACKED_SETTINGS.items():
        for setting_name in setting_names_list:
            if bot.memory[SETTING_CACHE_NAMESPACE][section][setting_name] != getattr(getattr(bot.config, section), setting_name):
                return False
    return True


@plugin.rule('$nick' r'(?i)(help|doc) +([A-Za-z]+)(?:\?+)?$')
@plugin.example('.help tell')
@plugin.command('help', 'commands')
@plugin.priority('low')
def help(bot, trigger):
    """Shows a command's documentation, and an example if available. With no arguments, lists all commands."""
    if bot.config.help.reply_method == 'query':
        def respond(text):
            bot.say(text, trigger.nick)
    elif bot.config.help.reply_method == 'notice':
        def respond(text):
            bot.notice(text, trigger.nick)
    else:
        def respond(text):
            bot.say(text, trigger.sender)

    if trigger.group(2):
        name = trigger.group(2)
        name = name.lower()

        # number of lines of help to show
        threshold = 3

        if name in bot.doc:
            # count lines we're going to send
            # lines in command docstring, plus one line for example(s) if present (they're sent all on one line)
            if len(bot.doc[name][0]) + int(bool(bot.doc[name][1])) > threshold:
                if trigger.nick != trigger.sender:  # don't say that if asked in private
                    bot.reply('The documentation for this command is too long; '
                              'I\'m sending it to you in a private message.')

                def msgfun(l):
                    bot.say(l, trigger.nick)
            else:
                msgfun = respond

            for line in bot.doc[name][0]:
                msgfun(line)
            if bot.doc[name][1]:
                # Build a nice, grammatically-correct list of examples
                examples = ', '.join(bot.doc[name][1][:-2] + [' or '.join(bot.doc[name][1][-2:])])
                msgfun('e.g. ' + examples)
    else:
        # This'll probably catch most cases, without having to spend the time
        # actually creating the list first. Maybe worth storing the link and a
        # heuristic in the DB, too, so it persists across restarts. Would need a
        # command to regenerate, too...
        if (
            'command-list' in bot.memory and
            bot.memory['command-list'][0] == len(bot.command_groups) and
            is_cache_valid(bot)
        ):
            url = bot.memory['command-list'][1]
        else:
            respond("Hang on, I'm creating a list.")
            msgs = []

            name_length = max(6, max(len(k) for k in bot.command_groups.keys()))
            for category, cmds in collections.OrderedDict(sorted(bot.command_groups.items())).items():
                category = category.upper().ljust(name_length)
                cmds = set(cmds)  # remove duplicates
                cmds = '  '.join(cmds)
                msg = category + '  ' + cmds
                indent = ' ' * (name_length + 2)
                # Honestly not sure why this is a list here
                msgs.append('\n'.join(textwrap.wrap(msg, subsequent_indent=indent)))

            url = create_list(bot, '\n\n'.join(msgs))
            if not url:
                return
            bot.memory['command-list'] = (len(bot.command_groups), url)
            update_cache(bot)
        respond("I've posted a list of my commands at {0} - You can see "
                "more info about any of these commands by doing {1}help "
                "<command> (e.g. {1}help time)"
                .format(url, bot.config.core.help_prefix))


def create_list(bot, msg):
    """Creates & uploads the command list.

    Returns the URL from the chosen pastebin provider.
    """
    msg = 'Command listing for {}{}\n\n{}'.format(
        bot.nick,
        ('@' + bot.config.core.host) if bot.config.help.show_server_host else '',
        msg)

    try:
        result = PASTEBIN_PROVIDERS[bot.config.help.output](msg)
    except PostingException:
        bot.say("Sorry! Something went wrong.")
        LOGGER.exception("Error posting commands")
        return
    return result


@plugin.rule('$nick' r'(?i)help(?:[?!]+)?$')
@plugin.priority('low')
def help2(bot, trigger):
    response = (
        "Hi, I'm a bot. Say {1}commands to me in private for a list "
        "of my commands, or see https://sopel.chat for more "
        "general details. My owner is {0}."
        .format(bot.config.core.owner, bot.config.core.help_prefix))
    bot.reply(response)



# coding=utf-8
"""Tests for message parsing"""
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import re

import pytest

from sopel.tools import Identifier
from sopel.trigger import PreTrigger, Trigger


TMP_CONFIG = """
[core]
owner = Foo
admins =
    Bar
"""


TMP_CONFIG_ACCOUNT = """
[core]
owner = Foo
owner_account = bar
admins =
    Bar
"""


@pytest.fixture
def nick():
    return Identifier('Sopel')


def test_basic_pretrigger(nick):
    line = ':Foo!foo@example.com PRIVMSG #Sopel :Hello, world'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.tags == {}
    assert pretrigger.hostmask == 'Foo!foo@example.com'
    assert pretrigger.line == line
    assert pretrigger.args == ['#Sopel', 'Hello, world']
    assert pretrigger.text == 'Hello, world'
    assert pretrigger.plain == 'Hello, world'
    assert pretrigger.event == 'PRIVMSG'
    assert pretrigger.nick == Identifier('Foo')
    assert pretrigger.user == 'foo'
    assert pretrigger.host == 'example.com'
    assert pretrigger.sender == '#Sopel'


def test_pm_pretrigger(nick):
    line = ':Foo!foo@example.com PRIVMSG Sopel :Hello, world'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.tags == {}
    assert pretrigger.hostmask == 'Foo!foo@example.com'
    assert pretrigger.line == line
    assert pretrigger.args == ['Sopel', 'Hello, world']
    assert pretrigger.text == 'Hello, world'
    assert pretrigger.plain == 'Hello, world'
    assert pretrigger.event == 'PRIVMSG'
    assert pretrigger.nick == Identifier('Foo')
    assert pretrigger.user == 'foo'
    assert pretrigger.host == 'example.com'
    assert pretrigger.sender == Identifier('Foo')


def test_quit_pretrigger(nick):
    line = ':Foo!foo@example.com QUIT :quit message text'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.tags == {}
    assert pretrigger.hostmask == 'Foo!foo@example.com'
    assert pretrigger.line == line
    assert pretrigger.args == ['quit message text']
    assert pretrigger.text == 'quit message text'
    assert pretrigger.plain == 'quit message text'
    assert pretrigger.event == 'QUIT'
    assert pretrigger.nick == Identifier('Foo')
    assert pretrigger.user == 'foo'
    assert pretrigger.host == 'example.com'
    assert pretrigger.sender is None


def test_join_pretrigger(nick):
    line = ':Foo!foo@example.com JOIN #Sopel'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.tags == {}
    assert pretrigger.hostmask == 'Foo!foo@example.com'
    assert pretrigger.line == line
    assert pretrigger.args == ['#Sopel']
    assert pretrigger.text == '#Sopel'
    assert pretrigger.plain == '#Sopel'
    assert pretrigger.event == 'JOIN'
    assert pretrigger.nick == Identifier('Foo')
    assert pretrigger.user == 'foo'
    assert pretrigger.host == 'example.com'
    assert pretrigger.sender == Identifier('#Sopel')


def test_tags_pretrigger(nick):
    line = '@foo=bar;baz;sopel.chat/special=value :Foo!foo@example.com PRIVMSG #Sopel :Hello, world'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.tags == {'baz': None,
                               'foo': 'bar',
                               'sopel.chat/special': 'value'}
    assert pretrigger.hostmask == 'Foo!foo@example.com'
    assert pretrigger.line == line
    assert pretrigger.args == ['#Sopel', 'Hello, world']
    assert pretrigger.text == 'Hello, world'
    assert pretrigger.plain == 'Hello, world'
    assert pretrigger.event == 'PRIVMSG'
    assert pretrigger.nick == Identifier('Foo')
    assert pretrigger.user == 'foo'
    assert pretrigger.host == 'example.com'
    assert pretrigger.sender == '#Sopel'


def test_intents_pretrigger(nick):
    line = '@intent=ACTION :Foo!foo@example.com PRIVMSG #Sopel :Hello, world'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.tags == {'intent': 'ACTION'}
    assert pretrigger.hostmask == 'Foo!foo@example.com'
    assert pretrigger.line == line
    assert pretrigger.args == ['#Sopel', 'Hello, world']
    assert pretrigger.text == 'Hello, world'
    assert pretrigger.plain == 'Hello, world'
    assert pretrigger.event == 'PRIVMSG'
    assert pretrigger.nick == Identifier('Foo')
    assert pretrigger.user == 'foo'
    assert pretrigger.host == 'example.com'
    assert pretrigger.sender == '#Sopel'


def test_unusual_pretrigger(nick):
    line = 'PING'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.tags == {}
    assert pretrigger.hostmask is None
    assert pretrigger.line == line
    assert pretrigger.args == []
    assert pretrigger.text == 'PING'
    assert pretrigger.plain == ''
    assert pretrigger.event == 'PING'


def test_ctcp_intent_pretrigger(nick):
    line = ':Foo!foo@example.com PRIVMSG Sopel :\x01VERSION\x01'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.tags == {'intent': 'VERSION'}
    assert pretrigger.hostmask == 'Foo!foo@example.com'
    assert pretrigger.line == line
    assert pretrigger.args == ['Sopel', '']
    assert pretrigger.text == '\x01VERSION\x01'
    assert pretrigger.plain == ''
    assert pretrigger.event == 'PRIVMSG'
    assert pretrigger.nick == Identifier('Foo')
    assert pretrigger.user == 'foo'
    assert pretrigger.host == 'example.com'
    assert pretrigger.sender == Identifier('Foo')


def test_ctcp_data_pretrigger(nick):
    line = ':Foo!foo@example.com PRIVMSG Sopel :\x01PING 1123321\x01'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.tags == {'intent': 'PING'}
    assert pretrigger.hostmask == 'Foo!foo@example.com'
    assert pretrigger.line == line
    assert pretrigger.args == ['Sopel', '1123321']
    assert pretrigger.text == '\x01PING 1123321\x01'
    assert pretrigger.plain == '1123321'
    assert pretrigger.event == 'PRIVMSG'
    assert pretrigger.nick == Identifier('Foo')
    assert pretrigger.user == 'foo'
    assert pretrigger.host == 'example.com'
    assert pretrigger.sender == Identifier('Foo')


def test_ctcp_action_pretrigger(nick):
    line = ':Foo!foo@example.com PRIVMSG #Sopel :\x01ACTION Hello, world\x01'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.tags == {'intent': 'ACTION'}
    assert pretrigger.hostmask == 'Foo!foo@example.com'
    assert pretrigger.line == line
    assert pretrigger.args == ['#Sopel', 'Hello, world']
    assert pretrigger.text == '\x01ACTION Hello, world\x01'
    assert pretrigger.plain == 'Hello, world'
    assert pretrigger.event == 'PRIVMSG'
    assert pretrigger.nick == Identifier('Foo')
    assert pretrigger.user == 'foo'
    assert pretrigger.host == 'example.com'
    assert pretrigger.sender == '#Sopel'


def test_ctcp_action_trigger(nick, configfactory):
    line = ':Foo!bar@example.com PRIVMSG #Sopel :\x01ACTION Hello, world\x01'
    pretrigger = PreTrigger(nick, line)

    config = configfactory('default.cfg', TMP_CONFIG)
    fakematch = re.match('.*', line)

    trigger = Trigger(config, pretrigger, fakematch)
    assert trigger.sender == '#Sopel'
    assert trigger.raw == line
    assert trigger.is_privmsg is False
    assert trigger.hostmask == 'Foo!bar@example.com'
    assert trigger.user == 'bar'
    assert trigger.nick == Identifier('Foo')
    assert trigger.host == 'example.com'
    assert trigger.event == 'PRIVMSG'
    assert trigger.match == fakematch
    assert trigger.group == fakematch.group
    assert trigger.groups == fakematch.groups
    assert trigger.groupdict == fakematch.groupdict
    assert trigger.args == ['#Sopel', 'Hello, world']
    assert trigger.plain == 'Hello, world'
    assert trigger.tags == {'intent': 'ACTION'}
    assert trigger.ctcp == 'ACTION'
    assert trigger.account is None
    assert trigger.admin is True
    assert trigger.owner is True


def test_ircv3_extended_join_pretrigger(nick):
    line = ':Foo!foo@example.com JOIN #Sopel bar :Real Name'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.tags == {'account': 'bar'}
    assert pretrigger.hostmask == 'Foo!foo@example.com'
    assert pretrigger.line == line
    assert pretrigger.args == ['#Sopel', 'bar', 'Real Name']
    assert pretrigger.text == 'Real Name'
    assert pretrigger.plain == 'Real Name'
    assert pretrigger.event == 'JOIN'
    assert pretrigger.nick == Identifier('Foo')
    assert pretrigger.user == 'foo'
    assert pretrigger.host == 'example.com'
    assert pretrigger.sender == Identifier('#Sopel')


def test_ircv3_extended_join_trigger(nick, configfactory):
    line = ':Foo!foo@example.com JOIN #Sopel bar :Real Name'
    pretrigger = PreTrigger(nick, line)

    config = configfactory('default.cfg', TMP_CONFIG)

    fakematch = re.match('.*', line)

    trigger = Trigger(config, pretrigger, fakematch)
    assert trigger.sender == '#Sopel'
    assert trigger.raw == line
    assert trigger.is_privmsg is False
    assert trigger.hostmask == 'Foo!foo@example.com'
    assert trigger.user == 'foo'
    assert trigger.nick == Identifier('Foo')
    assert trigger.host == 'example.com'
    assert trigger.event == 'JOIN'
    assert trigger.match == fakematch
    assert trigger.group == fakematch.group
    assert trigger.groups == fakematch.groups
    assert trigger.args == ['#Sopel', 'bar', 'Real Name']
    assert trigger.plain == 'Real Name'
    assert trigger.account == 'bar'
    assert trigger.tags == {'account': 'bar'}
    assert trigger.ctcp is None
    assert trigger.owner is True
    assert trigger.admin is True


def test_ircv3_intents_trigger(nick, configfactory):
    line = '@intent=ACTION :Foo!bar@example.com PRIVMSG #Sopel :Hello, world'
    pretrigger = PreTrigger(nick, line)

    config = configfactory('default.cfg', TMP_CONFIG)
    fakematch = re.match('.*', line)

    trigger = Trigger(config, pretrigger, fakematch)
    assert trigger.sender == '#Sopel'
    assert trigger.raw == line
    assert trigger.is_privmsg is False
    assert trigger.hostmask == 'Foo!bar@example.com'
    assert trigger.user == 'bar'
    assert trigger.nick == Identifier('Foo')
    assert trigger.host == 'example.com'
    assert trigger.event == 'PRIVMSG'
    assert trigger.match == fakematch
    assert trigger.group == fakematch.group
    assert trigger.groups == fakematch.groups
    assert trigger.groupdict == fakematch.groupdict
    assert trigger.args == ['#Sopel', 'Hello, world']
    assert trigger.plain == 'Hello, world'
    assert trigger.tags == {'intent': 'ACTION'}
    assert trigger.ctcp == 'ACTION'
    assert trigger.account is None
    assert trigger.admin is True
    assert trigger.owner is True


def test_ircv3_account_tag_trigger(nick, configfactory):
    line = '@account=bar :Nick_Is_Not_Foo!foo@example.com PRIVMSG #Sopel :Hello, world'
    pretrigger = PreTrigger(nick, line)

    config = configfactory('default.cfg', TMP_CONFIG_ACCOUNT)
    fakematch = re.match('.*', line)

    trigger = Trigger(config, pretrigger, fakematch)
    assert trigger.admin is True
    assert trigger.owner is True


def test_ircv3_server_time_trigger(nick, configfactory):
    line = '@time=2016-01-09T03:15:42.000Z :Foo!foo@example.com PRIVMSG #Sopel :Hello, world'
    pretrigger = PreTrigger(nick, line)
    config = configfactory('default.cfg', TMP_CONFIG)
    fakematch = re.match('.*', line)

    trigger = Trigger(config, pretrigger, fakematch)
    assert trigger.time == datetime.datetime(2016, 1, 9, 3, 15, 42, 0)

    # Spec-breaking string
    line = '@time=2016-01-09T04:20 :Foo!foo@example.com PRIVMSG #Sopel :Hello, world'
    pretrigger = PreTrigger(nick, line)
    assert pretrigger.time is not None



# coding=utf-8
"""
tell.py - Sopel Tell and Ask Plugin
Copyright 2008, Sean B. Palmer, inamidst.com
Copyright 2019, dgw, technobabbl.es
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from collections import defaultdict
import io  # don't use `codecs` for loading the DB; it will split lines on some IRC formatting
import logging
import os
import sys
import threading
import time
import unicodedata

from sopel import formatting, plugin, tools
from sopel.config import types
from sopel.tools.time import format_time, get_timezone


if sys.version_info.major >= 3:
    unicode = str


LOGGER = logging.getLogger(__name__)


class TellSection(types.StaticSection):
    use_private_reminder = types.BooleanAttribute(
        'use_private_reminder', default=False)
    """When set to ``true``, Sopel will send reminder as private message."""
    maximum_public = types.ValidatedAttribute(
        'maximum_public', parse=int, default=4)
    """How many Sopel can send in public before using private message."""


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | use_private_reminder | false | Send reminders as private message |
    | maximum_public | 4 | Send up to this amount of reminders in public |
    """
    config.define_section('tell', TellSection)
    config.tell.configure_setting(
        'use_private_reminder',
        'Should Sopel send tell/ask reminders as private message only?')
    if not config.tell.use_private_reminder:
        config.tell.configure_setting(
            'maximum_public',
            'How many tell/ask reminders Sopel will send as public message '
            'before sending them as private messages?')


def load_reminders(filename):
    """Load tell/ask reminders from a ``filename``.

    :param str filename: path to the tell/ask reminders file
    :return: a dict with the tell/ask reminders
    :rtype: dict
    """
    result = defaultdict(list)
    with io.open(filename, 'r', encoding='utf-8') as fd:
        for line in fd:
            line = line.strip()
            if line:
                try:
                    tellee, teller, verb, timenow, msg = line.split('\t', 4)
                except ValueError:
                    continue  # TODO: Add warning log about malformed reminder
                result[tellee].append((teller, verb, timenow, msg))

    return result


def dump_reminders(filename, data):
    """Dump tell/ask reminders (``data``) into a ``filename``.

    :param str filename: path to the tell/ask reminders file
    :param dict data: tell/ask reminders ``dict``
    """
    with io.open(filename, 'w', encoding='utf-8') as fd:
        for tellee, reminders in data.items():
            for reminder in reminders:
                line = '\t'.join((tellee,) + tuple(reminder))
                fd.write(line + '\n')
    return True


def setup(bot):
    bot.config.define_section('tell', TellSection)
    fn = bot.config.basename + '.tell.db'
    bot.tell_filename = os.path.join(bot.config.core.homedir, fn)

    # Pre-7.0 migration logic. Remove in 8.0 or 9.0.
    old = bot.nick + '-' + bot.config.core.host + '.tell.db'
    old = os.path.join(bot.config.core.homedir, old)
    if os.path.isfile(old):
        LOGGER.info("Attempting to migrate old 'tell' database {}..."
                    .format(old))
        try:
            os.rename(old, bot.tell_filename)
        except OSError:
            LOGGER.error("Migration failed!")
            LOGGER.error("Old filename: {}".format(old))
            LOGGER.error("New filename: {}".format(bot.tell_filename))
            LOGGER.error(
                "See https://sopel.chat/usage/installing/upgrading-to-sopel-7/#reminder-db-migration")
        else:
            LOGGER.info("Migration finished!")
    # End migration logic

    if not os.path.exists(bot.tell_filename):
        with io.open(bot.tell_filename, 'w', encoding='utf-8') as fd:
            # if we can't open/write into the file, the tell plugin can't work
            fd.write('')

    if 'tell_lock' not in bot.memory:
        bot.memory['tell_lock'] = threading.Lock()

    if 'reminders' not in bot.memory:
        with bot.memory['tell_lock']:
            bot.memory['reminders'] = load_reminders(bot.tell_filename)


def shutdown(bot):
    for key in ['tell_lock', 'reminders']:
        try:
            del bot.memory[key]
        except KeyError:
            pass


def _format_safe_lstrip(text):
    """``str.lstrip()`` but without eating IRC formatting.

    :param str text: text to clean
    :rtype: str
    :raises TypeError: if the passed ``text`` is not a string

    Stolen and tweaked from the ``choose`` plugin's ``_format_safe()``
    function by the person who wrote it.
    """
    if not isinstance(text, unicode):
        raise TypeError("A string is required.")
    elif not text:
        # unnecessary optimization
        return ''

    start = 0

    # strip left
    pos = 0
    while pos < len(text):
        is_whitespace = unicodedata.category(text[pos]) == 'Zs'
        is_non_printing = (
            text[pos] in formatting.CONTROL_NON_PRINTING and
            text[pos] not in formatting.CONTROL_FORMATTING
        )
        if not is_whitespace and not is_non_printing:
            start = pos
            break
        pos += 1
    else:
        # skipped everything; string is all whitespace
        return ''

    return text[start:]


@plugin.command('tell', 'ask')
@plugin.nickname_command('tell', 'ask')
@plugin.example('$nickname, tell dgw he broke something again.')
def f_remind(bot, trigger):
    """Give someone a message the next time they're seen"""
    teller = trigger.nick
    verb = trigger.group(1)

    if not trigger.group(3):
        bot.reply("%s whom?" % verb)
        return

    tellee = trigger.group(3).rstrip('.,:;')
    msg = _format_safe_lstrip(trigger.group(2).split(' ', 1)[1])

    if not msg:
        bot.reply("%s %s what?" % (verb, tellee))
        return

    tellee = tools.Identifier(tellee)

    if not os.path.exists(bot.tell_filename):
        return

    if len(tellee) > bot.isupport.get('NICKLEN', 30):
        bot.reply('That nickname is too long.')
        return

    if tellee[0] == '@':
        tellee = tellee[1:]

    if tellee == bot.nick:
        bot.reply("I'm here now; you can %s me whatever you want!" % verb)
        return

    if tellee not in (tools.Identifier(teller), bot.nick, 'me'):
        tz = get_timezone(bot.db, bot.config, None, tellee)
        timenow = format_time(bot.db, bot.config, tz, tellee)
        with bot.memory['tell_lock']:
            if tellee not in bot.memory['reminders']:
                bot.memory['reminders'][tellee] = [(teller, verb, timenow, msg)]
            else:
                bot.memory['reminders'][tellee].append((teller, verb, timenow, msg))
            # save the reminders
            dump_reminders(bot.tell_filename, bot.memory['reminders'])

        response = "I'll pass that on when %s is around." % tellee
        bot.reply(response)
    elif tools.Identifier(teller) == tellee:
        bot.reply('You can %s yourself that.' % verb)
    else:
        bot.reply("Hey, I'm not as stupid as Monty you know!")


def get_nick_reminders(reminders, nick):
    lines = []
    template = "%s: %s <%s> %s %s %s"
    today = time.strftime('%d %b', time.gmtime())

    for (teller, verb, datetime, msg) in reminders:
        if datetime.startswith(today):
            datetime = datetime[len(today) + 1:]
        lines.append(template % (nick, datetime, teller, verb, nick, msg))

    return lines


def nick_match_tellee(nick, tellee):
    """Tell if a ``nick`` matches a ``tellee``.

    :param str nick: Nick seen by the bot
    :param str tellee: Tellee name or pattern

    The check between ``nick`` and ``tellee`` is case-insensitive::

        >>> nick_match_tellee('Exirel', 'exirel')
        True
        >>> nick_match_tellee('exirel', 'EXIREL')
        True
        >>> nick_match_tellee('exirel', 'dgw')
        False

    If ``tellee`` ends with a wildcard token (``*`` or ``:``), then ``nick``
    matches if it starts with ``tellee`` (without the token)::

        >>> nick_match_tellee('Exirel', 'Exi*')
        True
        >>> nick_match_tellee('Exirel', 'exi:')
        True
        >>> nick_match_tellee('Exirel', 'Exi')
        False

    Note that this is still case-insensitive.
    """
    if tellee[-1] in ['*', ':']:  # these are wildcard token
        return nick.lower().startswith(tellee.lower().rstrip('*:'))
    return nick.lower() == tellee.lower()


@plugin.rule('(.*)')
@plugin.priority('low')
@plugin.unblockable
@plugin.output_prefix('[tell] ')
def message(bot, trigger):
    nick = trigger.nick

    if not os.path.exists(bot.tell_filename):
        # plugin can't work without its storage file
        return

    # get all matching reminders
    reminders = []
    tellees = list(reversed(sorted(
        tellee
        for tellee in bot.memory['reminders']
        if nick_match_tellee(nick, tellee)
    )))

    with bot.memory['tell_lock']:
        # pop reminders for nick
        reminders = list(
            reminder
            for tellee in tellees
            for reminder in get_nick_reminders(
                bot.memory['reminders'].pop(tellee, []), nick)
        )

    # check if there are reminders to send
    if not reminders:
        return  # nothing to do

    # then send reminders (as public and/or private messages)
    if bot.config.tell.use_private_reminder:
        # send reminders with private messages
        for line in reminders:
            bot.say(line, nick)
    else:
        # send up to 'maximum_public' reminders to the channel
        max_public = bot.config.tell.maximum_public
        for line in reminders[:max_public]:
            bot.say(line)

        # send other reminders directly to nick as private message
        if reminders[max_public:]:
            bot.reply('Further messages sent privately')
            for line in reminders[max_public:]:
                bot.say(line, nick)

    # save reminders left in memory
    with bot.memory['tell_lock']:
        dump_reminders(bot.tell_filename, bot.memory['reminders'])



# coding=utf-8
"""
py.py - Sopel Python Eval Plugin
Copyright 2008, Sean B. Palmer, inamidst.com
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from requests import get

from sopel import plugin
from sopel.config import types
from sopel.tools.web import quote


class PySection(types.StaticSection):
    oblique_instance = types.ValidatedAttribute(
        'oblique_instance',
        default='https://oblique.sopel.chat/')
    """The Oblique instance to use when evaluating Python expressions"""


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | oblique_instance | https://oblique.sopel.chat/ | The Oblique instance to use when evaluating Python expressions (see <https://github.com/sopel-irc/oblique>) |
    """
    config.define_section('py', PySection)
    config.py.configure_setting(
        'oblique_instance',
        'Enter the base URL of a custom Oblique instance (optional): '
    )


def setup(bot):
    bot.config.define_section('py', PySection)

    if not any(
        bot.config.py.oblique_instance.startswith(prot)
        for prot in ['http://', 'https://']
    ):
        raise ValueError('Oblique instance URL must start with a protocol.')

    if not bot.config.py.oblique_instance.endswith('/'):
        bot.config.py.oblique_instance += '/'


@plugin.command('py')
@plugin.output_prefix('[py] ')
@plugin.example('.py len([1,2,3])', '3', online=True, vcr=True)
def py(bot, trigger):
    """Evaluate a Python expression."""
    query = trigger.group(2)
    if not query:
        bot.reply('What expression do you want me to evaluate?')
        return

    uri = bot.config.py.oblique_instance + 'py/'
    answer = get(uri + quote(query)).content.decode('utf-8')
    if answer:
        bot.say(answer)
    else:
        bot.reply('Sorry, no result.')



# coding=utf-8
"""Tests for sopel.plugin decorators"""
from __future__ import absolute_import, division, print_function, unicode_literals

from sopel import plugin
from sopel.tests import rawlist


TMP_CONFIG = """
[core]
owner = Bar
nick = Sopel
enable = coretasks
"""


def test_find():
    @plugin.find('.*')
    def mock(bot, trigger, match):
        return True
    assert mock.find_rules == ['.*']


def test_find_args():
    @plugin.find('.*', r'\d+')
    def mock(bot, trigger, match):
        return True
    assert mock.find_rules == ['.*', r'\d+']


def test_find_multiple():
    @plugin.find('.*', r'\d+')
    @plugin.find('.*')
    @plugin.find(r'\w+')
    def mock(bot, trigger, match):
        return True
    assert mock.find_rules == [r'\w+', '.*', r'\d+']


def test_label():
    @plugin.label('hello')
    def mock(bot, trigger):
        return True
    assert mock.rule_label == 'hello'


def test_search():
    @plugin.search('.*')
    def mock(bot, trigger, match):
        return True
    assert mock.search_rules == ['.*']


def test_search_args():
    @plugin.search('.*', r'\d+')
    def mock(bot, trigger, match):
        return True
    assert mock.search_rules == ['.*', r'\d+']


def test_search_multiple():
    @plugin.search('.*', r'\d+')
    @plugin.search('.*')
    @plugin.search(r'\w+')
    def mock(bot, trigger, match):
        return True
    assert mock.search_rules == [r'\w+', '.*', r'\d+']


def test_url_lazy():
    def loader(settings):
        return [r'\w+', '.*', r'\d+']

    @plugin.url_lazy(loader)
    def mock(bot, trigger, match):
        return True

    assert mock.url_lazy_loaders == [loader]
    assert not hasattr(mock, 'url_regex')


def test_url_lazy_args():
    def loader_1(settings):
        return [r'\w+', '.*', r'\d+']

    def loader_2(settings):
        return [r'[a-z]+']

    @plugin.url_lazy(loader_1, loader_2)
    def mock(bot, trigger, match):
        return True

    assert mock.url_lazy_loaders == [loader_1, loader_2]
    assert not hasattr(mock, 'url_regex')


def test_url_lazy_multiple():
    def loader_1(settings):
        return [r'\w+', '.*', r'\d+']

    def loader_2(settings):
        return [r'[a-z]+']

    @plugin.url_lazy(loader_2)
    @plugin.url_lazy(loader_1)
    def mock(bot, trigger, match):
        return True

    assert mock.url_lazy_loaders == [loader_1, loader_2]
    assert not hasattr(mock, 'url_regex')


def test_rule_lazy():
    def loader(settings):
        return [r'\w+', '.*', r'\d+']

    @plugin.rule_lazy(loader)
    def mock(bot, trigger):
        return True

    assert mock.rule_lazy_loaders == [loader]
    assert not hasattr(mock, 'rule')


def test_rule_lazy_args():
    def loader_1(settings):
        return [r'\w+', '.*', r'\d+']

    def loader_2(settings):
        return [r'[a-z]+']

    @plugin.rule_lazy(loader_1, loader_2)
    def mock(bot, trigger):
        return True

    assert mock.rule_lazy_loaders == [loader_1, loader_2]
    assert not hasattr(mock, 'rule')


def test_rule_lazy_multiple():
    def loader_1(settings):
        return [r'\w+', '.*', r'\d+']

    def loader_2(settings):
        return [r'[a-z]+']

    @plugin.rule_lazy(loader_2)
    @plugin.rule_lazy(loader_1)
    def mock(bot, trigger):
        return True

    assert mock.rule_lazy_loaders == [loader_1, loader_2]
    assert not hasattr(mock, 'rule')


def test_find_lazy():
    def loader(settings):
        return [r'\w+', '.*', r'\d+']

    @plugin.find_lazy(loader)
    def mock(bot, trigger):
        return True

    assert mock.find_rules_lazy_loaders == [loader]
    assert not hasattr(mock, 'rule')
    assert not hasattr(mock, 'find_rules')


def test_find_lazy_args():
    def loader_1(settings):
        return [r'\w+', '.*', r'\d+']

    def loader_2(settings):
        return [r'[a-z]+']

    @plugin.find_lazy(loader_1, loader_2)
    def mock(bot, trigger):
        return True

    assert mock.find_rules_lazy_loaders == [loader_1, loader_2]
    assert not hasattr(mock, 'rule')
    assert not hasattr(mock, 'find_rules')


def test_find_lazy_multiple():
    def loader_1(settings):
        return [r'\w+', '.*', r'\d+']

    def loader_2(settings):
        return [r'[a-z]+']

    @plugin.find_lazy(loader_2)
    @plugin.find_lazy(loader_1)
    def mock(bot, trigger):
        return True

    assert mock.find_rules_lazy_loaders == [loader_1, loader_2]
    assert not hasattr(mock, 'rule')
    assert not hasattr(mock, 'find_rules')


def test_search_lazy():
    def loader(settings):
        return [r'\w+', '.*', r'\d+']

    @plugin.search_lazy(loader)
    def mock(bot, trigger):
        return True

    assert mock.search_rules_lazy_loaders == [loader]
    assert not hasattr(mock, 'rule')
    assert not hasattr(mock, 'search_rules')


def test_search_lazy_args():
    def loader_1(settings):
        return [r'\w+', '.*', r'\d+']

    def loader_2(settings):
        return [r'[a-z]+']

    @plugin.search_lazy(loader_1, loader_2)
    def mock(bot, trigger):
        return True

    assert mock.search_rules_lazy_loaders == [loader_1, loader_2]
    assert not hasattr(mock, 'rule')
    assert not hasattr(mock, 'search_rules')


def test_search_lazy_multiple():
    def loader_1(settings):
        return [r'\w+', '.*', r'\d+']

    def loader_2(settings):
        return [r'[a-z]+']

    @plugin.search_lazy(loader_2)
    @plugin.search_lazy(loader_1)
    def mock(bot, trigger):
        return True

    assert mock.search_rules_lazy_loaders == [loader_1, loader_2]
    assert not hasattr(mock, 'rule')
    assert not hasattr(mock, 'search_rules')


def test_ctcp():
    @plugin.ctcp('ACTION')
    def mock(bot, trigger, match):
        return True
    assert mock.intents == ['ACTION']


def test_ctcp_empty():
    @plugin.ctcp()
    def mock(bot, trigger, match):
        return True
    assert mock.intents == ['ACTION']


def test_ctcp_direct():
    @plugin.ctcp
    def mock(bot, trigger, match):
        return True
    assert mock.intents == ['ACTION']


BAN_MESSAGE = ':Foo!foo@example.com PRIVMSG #chan :.ban ExiClone'
BAN_PRIVATE_MESSAGE = ':Foo!foo@example.com PRIVMSG Sopel :.ban #chan ExiClone'


def test_require_bot_privilege(configfactory,
                               botfactory,
                               triggerfactory,
                               ircfactory):
    settings = configfactory('default.cfg', TMP_CONFIG)
    mockbot = botfactory.preloaded(settings)
    mockserver = ircfactory(mockbot)

    bot = triggerfactory.wrapper(mockbot, BAN_MESSAGE)
    mockserver.channel_joined('#chan')
    mockserver.join('Foo', '#chan')
    mockserver.mode_set('#chan', '+vo', ['Foo', bot.nick])

    @plugin.command('ban')
    @plugin.require_bot_privilege(plugin.VOICE)
    def mock(bot, trigger):
        return True

    assert mock(bot, bot._trigger) is True, (
        'Bot must meet the requirement when having a higher privilege level.')

    @plugin.command('ban')
    @plugin.require_bot_privilege(plugin.OP)
    def mock(bot, trigger):
        return True

    assert mock(bot, bot._trigger) is True

    @plugin.command('ban')
    @plugin.require_bot_privilege(plugin.OWNER)
    def mock(bot, trigger):
        return True

    assert mock(bot, bot._trigger) is not True
    assert not bot.backend.message_sent

    @plugin.command('ban')
    @plugin.require_bot_privilege(plugin.OWNER, message='Nope')
    def mock(bot, trigger):
        return True

    assert mock(bot, bot._trigger) is not True
    assert bot.backend.message_sent == rawlist('PRIVMSG #chan :Nope')

    @plugin.command('ban')
    @plugin.require_bot_privilege(plugin.OWNER, message='Nope', reply=True)
    def mock(bot, trigger):
        return True

    assert mock(bot, bot._trigger) is not True
    assert bot.backend.message_sent[1:] == rawlist('PRIVMSG #chan :Foo: Nope')


def test_require_bot_privilege_private_message(configfactory,
                                               botfactory,
                                               triggerfactory,
                                               ircfactory):
    settings = configfactory('default.cfg', TMP_CONFIG)
    mockbot = botfactory.preloaded(settings)
    mockserver = ircfactory(mockbot)

    bot = triggerfactory.wrapper(mockbot, BAN_PRIVATE_MESSAGE)
    mockserver.channel_joined('#chan')
    mockserver.join('Foo', '#chan')
    mockserver.mode_set('#chan', '+vo', ['Foo', bot.nick])

    @plugin.command('ban')
    @plugin.require_bot_privilege(plugin.VOICE)
    def mock(bot, trigger):
        return True

    assert mock(bot, bot._trigger) is True

    @plugin.command('ban')
    @plugin.require_bot_privilege(plugin.OP)
    def mock(bot, trigger):
        return True

    assert mock(bot, bot._trigger) is True

    @plugin.command('ban')
    @plugin.require_bot_privilege(plugin.OWNER)
    def mock(bot, trigger):
        return True

    assert mock(bot, bot._trigger) is True, (
        'There must not be privilege check for a private message.')
